import os
import datetime
from flask import Flask, render_template_string, request
from flask_babel import Babel
from flask_mail import Mail
from flask_sqlalchemy import SQLAlchemy
from flask_user import login_required, SQLAlchemyAdapter, UserManager, UserMixin
from flask_user import roles_required, confirm_email_required


app = Flask(__name__)
db = SQLAlchemy(app)                            # Initialize Flask-SQLAlchemy


# Define the User data model. Make sure to add flask_user UserMixin!!
class User(db.Model, UserMixin):
    id = db.Column(db.Integer, primary_key=True)

    # User authentication information
    username = db.Column(db.String(50), nullable=True, unique=True)
    password = db.Column(db.String(255), nullable=False, server_default='')

    # User email information
    email = db.Column(db.String(255), nullable=True, unique=True)
    confirmed_at = db.Column(db.DateTime())

    # User information
    active = db.Column('is_active', db.Boolean(), nullable=False, server_default='0')
    first_name = db.Column(db.String(100), nullable=False, server_default='')
    last_name = db.Column(db.String(100), nullable=False, server_default='')

    # Relationships
    roles = db.relationship('Role', secondary='user_roles',
            backref=db.backref('users', lazy='dynamic'))


# Define UserEmail DataModel.
class UserEmail(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(db.Integer, db.ForeignKey('user.id'))

    # User email information
    email = db.Column(db.String(255), nullable=True, unique=True)
    confirmed_at = db.Column(db.DateTime())
    is_primary = db.Column(db.Boolean(), nullable=False, default=False)

    # Relationship
    user = db.relationship('User', uselist=False)


class UserInvitation(db.Model):
    __tablename__ = 'user_invite'
    id = db.Column(db.Integer, primary_key=True)
    email = db.Column(db.String(255), nullable=False)
    # save the user of the invitee
    invited_by_user_id = db.Column(db.Integer, db.ForeignKey('user.id'))
    # token used for registration page to identify user registering
    token = db.Column(db.String(100), nullable=False, server_default='')


# Define the Role data model
class Role(db.Model):
    id = db.Column(db.Integer(), primary_key=True)
    name = db.Column(db.String(50), unique=True)


# Define the UserRoles data model
class UserRoles(db.Model):
    id = db.Column(db.Integer(), primary_key=True)
    user_id = db.Column(db.Integer(), db.ForeignKey('user.id', ondelete='CASCADE'))
    role_id = db.Column(db.Integer(), db.ForeignKey('role.id', ondelete='CASCADE'))


# Use a Class-based config to avoid needing a 2nd file
# os.getenv() enables configuration through OS environment variables
class ConfigClass(object):
    # Flask settings
    SECRET_KEY =              os.getenv('SECRET_KEY',       'THIS IS AN INSECURE SECRET')
    SQLALCHEMY_DATABASE_URI = os.getenv('DATABASE_URL',     'sqlite:///tst_app.sqlite')
    CSRF_ENABLED = True

    # Flask-Mail settings
    MAIL_USERNAME =           os.getenv('MAIL_USERNAME',        'email@example.com')
    MAIL_PASSWORD =           os.getenv('MAIL_PASSWORD',        'password')
    MAIL_DEFAULT_SENDER =     os.getenv('MAIL_DEFAULT_SENDER',  '"MyApp" <noreply@example.com>')
    MAIL_SERVER =             os.getenv('MAIL_SERVER',          'smtp.gmail.com')
    MAIL_PORT =           int(os.getenv('MAIL_PORT',            '465'))
    MAIL_USE_SSL =            os.getenv('MAIL_USE_SSL',         True)

    # Flask-User settings
    USER_APP_NAME        = "AppName"                # Used by email templates
    USER_ENABLE_USERNAME        = True
    USER_ENABLE_EMAIL           = True
    USER_ENABLE_CONFIRM_EMAIL   = True
    USER_ENABLE_INVITATION      = True


def init_app(app, test_config=None):                   # For automated tests
    # Setup Flask and read config from ConfigClass defined above
    app.config.from_object(__name__+'.ConfigClass')

    # Load local_settings.py if file exists         # For automated tests
    try: app.config.from_object('local_settings')
    except: pass

    # Load optional test_config                     # For automated tests
    if test_config:
        app.config.update(test_config)

    # Initialize Flask extensions
    babel = Babel(app)                              # Initialize Flask-Babel
    mail = Mail(app)                                # Initialize Flask-Mail

    # Reset all the database tables
    db.create_all()

    # Setup Flask-User
    db_adapter = SQLAlchemyAdapter(db,  User, UserInvitationClass=UserInvitation)
    user_manager = UserManager(db_adapter, app)

    # Create regular 'member' user
    if not User.query.filter(User.username=='member').first():
        user = User(username='member', email='member@example.com', active=True,
                password=user_manager.hash_password('Password1'), confirmed_at=datetime.datetime.utcnow())
        db.session.add(user)
        db.session.commit()

    # Create 'user007' user with 'secret' and 'agent' roles
    if not User.query.filter(User.username=='user007').first():
        user1 = User(username='user007', email='user007@example.com', active=True,
                password=user_manager.hash_password('Password1'))
        user1.roles.append(Role(name='secret'))
        user1.roles.append(Role(name='agent'))
        db.session.add(user1)
        db.session.commit()

    # The '/' page is accessible to anyone
    @app.route('/')
    def home_page():
        return render_template_string("""
            {% extends "base.html" %}
            {% block content %}
            <h2>{%trans%}Home Page{%endtrans%}</h2>
            <p><a href="{{ url_for('user.login') }}">{%trans%}Sign in{%endtrans%}</a></p>
            {% endblock %}
            """)

    # The '/profile' page requires a logged-in user
    @app.route('/user/profile')
    @login_required                                 # Use of @login_required decorator
    @confirm_email_required
    def user_profile_page():
        return render_template_string("""
            {% extends "base.html" %}
            {% block content %}
            <h2>{%trans%}Profile Page{%endtrans%}</h2>
            <p> {%trans%}Hello{%endtrans%}
                {{ current_user.username or current_user.email }},</p>
            <p> <a href="{{ url_for('user.change_username') }}">
                {%trans%}Change username{%endtrans%}</a></p>
            <p> <a href="{{ url_for('user.change_password') }}">
                {%trans%}Change password{%endtrans%}</a></p>
            <p> <a href="{{ url_for('user.invite') }}">
                {%trans%}Invite User{%endtrans%}</a></p>
            <p> <a href="{{ url_for('user.logout') }}?next={{ url_for('user.login') }}">
                {%trans%}Sign out{%endtrans%}</a></p>
            {% endblock %}
            """)

    # The '/special' page requires a user that has the 'special' AND ('sauce' OR 'agent') role.
    @app.route('/special')
    @roles_required('secret', ['sauce', 'agent'])   # Use of @roles_required decorator
    def special_page():
        return render_template_string("""
            {% extends "base.html" %}
            {% block content %}
            <h2>{%trans%}Special Page{%endtrans%}</h2>
            {% endblock %}
            """)

    # For testing only
    app.db = db
    app.UserEmailClass = UserEmail

    return app


# Start development web server
if __name__=='__main__':
    app = create_app()
    app.run(host='0.0.0.0', port=5555, debug=True)



import pytest
from auviewer import file as auvfile

@pytest.fixture
def f():
    f = auvfile.File(None, None, 'examples/sample_file.h5', None)
    yield f
    del f

def test_pattern_detection_high_threshold(f):

    expected = []
    patterns = f.detectPatterns(
        type='patterndetection',
        series='/series_4:value',
        thresholdlow=None,
        thresholdhigh=4,
        duration=30,
        persistence=.7,
        maxgap=50,
        expected_frequency=0,
        min_density=0,
        drop_values_below=None,
        drop_values_above=None,
        drop_values_between=None,
    )
    assert patterns == expected

    expected = [[284.0, 970.0], [2936.0, 3389.0]]
    patterns = f.detectPatterns(
        type='patterndetection',
        series='/series_4:value',
        thresholdlow=None,
        thresholdhigh=2.5,
        duration=30,
        persistence=.7,
        maxgap=50,
        expected_frequency=0,
        min_density=0,
        drop_values_below=None,
        drop_values_above=None,
        drop_values_between=None,
    )
    assert patterns == expected

def test_pattern_detection_low_threshold(f):

    expected = []
    patterns = f.detectPatterns(
        type='patterndetection',
        series='/series_4:value',
        thresholdlow=-4,
        thresholdhigh=None,
        duration=30,
        persistence=.7,
        maxgap=50,
        expected_frequency=0,
        min_density=0,
        drop_values_below=None,
        drop_values_above=None,
        drop_values_between=None,
    )
    assert patterns == expected

    expected = [[1604.0, 2189.0], [4371.0, 4509.0]]
    patterns = f.detectPatterns(
        type='patterndetection',
        series='/series_4:value',
        thresholdlow=-2.5,
        thresholdhigh=None,
        duration=30,
        persistence=.7,
        maxgap=50,
        expected_frequency=0,
        min_density=0,
        drop_values_below=None,
        drop_values_above=None,
        drop_values_between=None,
    )
    assert patterns == expected

def test_pattern_detection_high_and_low_threshold(f):

    expected = []
    patterns = f.detectPatterns(
        type='patterndetection',
        series='/series_4:value',
        thresholdlow=-4,
        thresholdhigh=4,
        duration=30,
        persistence=.7,
        maxgap=50,
        expected_frequency=0,
        min_density=0,
        drop_values_below=None,
        drop_values_above=None,
        drop_values_between=None,
    )
    assert patterns == expected

    expected = [[284.0, 970.0], [1604.0, 2189.0], [2936.0, 3389.0], [4371.0, 4509.0]]
    patterns = f.detectPatterns(
        type='patterndetection',
        series='/series_4:value',
        thresholdlow=-2.5,
        thresholdhigh=2.5,
        duration=30,
        persistence=.7,
        maxgap=50,
        expected_frequency=0,
        min_density=0,
        drop_values_below=None,
        drop_values_above=None,
        drop_values_between=None,
    )
    assert patterns == expected

def test_pattern_detection_high_threshold_drop_above(f):

    expected = []
    patterns = f.detectPatterns(
        type='patterndetection',
        series='/series_4:value',
        thresholdlow=None,
        thresholdhigh=2.5,
        duration=30,
        persistence=.7,
        maxgap=50,
        expected_frequency=0,
        min_density=0,
        drop_values_below=None,
        drop_values_above=2.5,
        drop_values_between=None,
    )
    assert patterns == expected

    expected = [[291.0, 321.0], [931.0, 960.0], [2943.0, 3030.0], [3348.0, 3376.0]]
    patterns = f.detectPatterns(
        type='patterndetection',
        series='/series_4:value',
        thresholdlow=None,
        thresholdhigh=2.5,
        duration=30,
        persistence=.7,
        maxgap=50,
        expected_frequency=0,
        min_density=0,
        drop_values_below=None,
        drop_values_above=2.6,
        drop_values_between=None,
    )
    assert patterns == expected

def test_pattern_detection_low_threshold_drop_below(f):

    expected = []
    patterns = f.detectPatterns(
        type='patterndetection',
        series='/series_4:value',
        thresholdlow=-2.5,
        thresholdhigh=None,
        duration=30,
        persistence=.7,
        maxgap=50,
        expected_frequency=0,
        min_density=0,
        drop_values_below=-2.5,
        drop_values_above=None,
        drop_values_between=None,
    )
    assert patterns == expected

    expected = [[1638.0, 1654.0], [2141.0, 2180.0], [4373.0, 4478.0]]
    patterns = f.detectPatterns(
        type='patterndetection',
        series='/series_4:value',
        thresholdlow=-2.5,
        thresholdhigh=None,
        duration=30,
        persistence=.7,
        maxgap=50,
        expected_frequency=0,
        min_density=0,
        drop_values_below=-2.6,
        drop_values_above=None,
        drop_values_between=None,
    )
    assert patterns == expected

def test_pattern_detection_high_and_low_thresholds_drop_above_and_below(f):

    expected = []
    patterns = f.detectPatterns(
        type='patterndetection',
        series='/series_4:value',
        thresholdlow=-2.5,
        thresholdhigh=2.5,
        duration=30,
        persistence=.7,
        maxgap=50,
        expected_frequency=0,
        min_density=0,
        drop_values_below=-2.5,
        drop_values_above=2.5,
        drop_values_between=None,
    )
    assert patterns == expected

    expected = [[291.0, 321.0], [931.0, 960.0], [1638.0, 1654.0], [2141.0, 2180.0], [2943.0, 3030.0], [3348.0, 3376.0], [4373.0, 4478.0]]
    patterns = f.detectPatterns(
        type='patterndetection',
        series='/series_4:value',
        thresholdlow=-2.5,
        thresholdhigh=2.5,
        duration=30,
        persistence=.7,
        maxgap=50,
        expected_frequency=0,
        min_density=0,
        drop_values_below=-2.6,
        drop_values_above=2.6,
        drop_values_between=None,
    )
    assert patterns == expected

def test_pattern_detection_high_and_low_thresholds_drop_between(f):

    expected = []
    patterns = f.detectPatterns(
        type='patterndetection',
        series='/series_4:value',
        thresholdlow=-2.5,
        thresholdhigh=2.5,
        duration=30,
        persistence=.7,
        maxgap=50,
        expected_frequency=0,
        min_density=0,
        drop_values_below=None,
        drop_values_above=None,
        drop_values_between=[-5, 5],
    )
    assert patterns == expected

    expected = [[262.0, 978.0], [1582.0, 2200.0], [2895.0, 3449.0], [4258.0, 4589.0], [4640.0, 4640.0], [5686.0, 5727.0]]
    patterns = f.detectPatterns(
        type='patterndetection',
        series='/series_4:value',
        thresholdlow=-2.5,
        thresholdhigh=2.5,
        duration=30,
        persistence=.7,
        maxgap=50,
        expected_frequency=0,
        min_density=0,
        drop_values_below=None,
        drop_values_above=None,
        drop_values_between=[-2.5, 2.5],
    )
    assert patterns == expected


""" This file implements Flask-User decorators: @login_required, @confirm_email_required and @roles_required.

    :copyright: (c) 2013 by Ling Thio
    :author: Ling Thio (ling.thio@gmail.com)
    :license: Simplified BSD License, see LICENSE.txt for more details."""

from functools import wraps
from flask import current_app
from flask_login import current_user
from .access import is_authenticated, is_confirmed_email


# Here to not break backward compatibility
def _call_or_get(function_or_property):
    return function_or_property() if callable(function_or_property) else function_or_property


def login_required(func):
    """ This decorator ensures that the current user is logged in before calling the actual view.
        Calls the unauthorized_view_function() when the user is not logged in."""
    @wraps(func)
    def decorated_view(*args, **kwargs):
        # User must be authenticated
        if not is_authenticated():
            # Redirect to unauthenticated page
            return current_app.user_manager.unauthenticated_view_function()

        # Call the actual view
        return func(*args, **kwargs)
    return decorated_view


def roles_accepted(*role_names):
    """ This decorator ensures that the current user one of the specified roles.
        Calls the unauthorized_view_function() when requirements fail.
        See also: UserMixin.has_role()
    """
    def wrapper(func):
        @wraps(func)
        def decorated_view(*args, **kwargs):
            # User must be logged
            if not is_authenticated():
                # Redirect to the unauthenticated page
                return current_app.user_manager.unauthenticated_view_function()

            # User must have the required roles
            if not current_user.has_role(*role_names):
                # Redirect to the unauthorized page
                return current_app.user_manager.unauthorized_view_function()

            # Call the actual view
            return func(*args, **kwargs)
        return decorated_view
    return wrapper


def roles_required(*role_names):
    """ This decorator ensures that the current user has all of the specified roles.
        Calls the unauthorized_view_function() when requirements fail.
        See also: UserMixin.has_roles()
    """
    def wrapper(func):
        @wraps(func)
        def decorated_view(*args, **kwargs):
            # User must be logged
            if not is_authenticated():
                # Redirect to the unauthenticated page
                return current_app.user_manager.unauthenticated_view_function()

            # User must have the required roles
            if not current_user.has_roles(*role_names):
                # Redirect to the unauthorized page
                return current_app.user_manager.unauthorized_view_function()

            # Call the actual view
            return func(*args, **kwargs)
        return decorated_view
    return wrapper


def confirm_email_required(func):
    """ This decorator ensures that the current user is logged in and has confirmed their email.
        Calls the unauthorized_view_function() when the user is not logged in."""
    @wraps(func)
    def decorated_view(*args, **kwargs):
        # User must be authenticated
        if is_authenticated():
            user_manager = current_app.user_manager
            # If confirm email has been enabled, user must have at least one confirmed email
            if is_confirmed_email():
                return func(*args, **kwargs)

        return current_app.user_manager.unconfirmed_email_view_function()

    return decorated_view



"""Python API for working with AUViewer."""

import sys
import logging
import traceback

from flask import Flask
from pathlib import Path
from typing import List, Dict, Optional
from pathlib import Path

from . import models
from .config import config, set_data_path
from .file import File
from .project import Project
from .shared import createEmptyJSONFile, getProcFNFromOrigFN

import multiprocessing as mp

# Will hold loaded projects
loadedProjects = []

# Initiate the donwsampling pool
downsamplePool = None

def downsampleFile(filepath: str, destinationpath: str) -> bool:
    """
    Downsamples an original file, placing the processed file in the destination folder.
    Raises an exception in case of error.
    :param filepath: path to the original file
    :param destinationpath: path to the destination folder
    :return: None
    """
    fp = Path(filepath)
    if not (fp.exists() and fp.is_file()):
        raise Exception(f"File '{filepath}' does not exist or is not a file.")

    dp = Path(destinationpath)
    if not (dp.exists() and dp.is_dir()):
        raise Exception(f"Destination '{destinationpath}' does not exist or is not a directory.")

    ds_file = File(None, -1, fp, dp / getProcFNFromOrigFN(fp))
    ds_file.process()
    ds_file.close()
    del ds_file

# Sets a global downsample
def instantiatePool(pool):
    global downsamplePool
    if downsamplePool is None: 
        downsamplePool = pool


def getProject(id) -> Optional[Project]:
    """
    Returns the project with matching ID.
    :return: the project instance belonging to the id, or None if not found
    """
    global loadedProjects
    for p in loadedProjects:
        if p.id == id:
            return p
    return None

def getProjects() -> Dict[int, Project]:
    """
    Returns loaded projects as a dict.
    :return: dict mapped from project ID to project instance of all loaded projects
    """
    global loadedProjects
    return {p.id: p for p in loadedProjects}

def getProjectsPayload(user_id) -> List[Dict]:
    """
    Returns a list of project information accessible to a given user.
    :return: list of objects containing project information
    """
    global loadedProjects
    return [{
        'id': p.id,
        'name': p.name,
        'files': len(p.files),
        'patterns': p.getTotalPatternCount(),
        'annotations': models.Annotation.query.filter_by(user_id=user_id, project_id=p.id).count(),
        'assignments_rem': models.User.query.filter_by(id=user_id).first().assignments_remaining,
    } for p in loadedProjects]

def listAvailableProjects() -> List[List[str]]:
    """
    Returns list of available projects (ID, name, path) from the database.
    These projects are not necessarily loaded in memory (that is done using
    loadProjects() or loadProject().
    :return:
    """
    return [[p.id, p.name, p.path] for p in models.Project.query.all()]

def listLoadedProjects() -> List[List[str]]:
    """
    Returns list of projects loaded in memory (ID, name, path).
    :return: list of lists
    """
    return [[p.id, p.name, p.path] for p in loadedProjects]

def listUsers() -> List[List[str]]:
    """
    Returns list of users (ID, email, first name, last name)
    :return: list of lists
    """
    return [[u.id, u.email, u.first_name, u.last_name] for u in models.User.query.all()]

def loadProject(id) -> Optional[Project]:
    """
    Load a project into memory. Returns the project instance if successful.
    :return: the project instance, or None
    """
    global loadedProjects

    logging.info(f"Loading project {id}")

    p = models.Project.query.filter_by(id=id).first()

    # Path object for the project
    projDirPathObj = Path(p.path)

    # Validate the project folder
    try:
        validateProjectFolder(projDirPathObj)
    except Exception as e:
        logging.error(f'Project folder {projDirPathObj} is invalid.\n{e}\n{traceback.format_exc()}')
        # TODO(gus): Update the database to reflect this?
    else:
        # TODO(gus): We need to have project take absolute path and project name!
        # Instantiate project, and add to the list to be returned
        loadedProjects.append(Project(p))

    logging.info("Finished loading project.")

    return getProject(id)

def loadProjects() -> Dict[int, Project]:
    """
    Load or reload projects into memory. If new projects are found on disk, they
    will be added to the database & loaded as well. Returns the same output as
    getProjects().
    :return: dict mapped from project ID to project of all loaded projects
    """

    global loadedProjects

    # Instantiate a global downsample pool
    pool = mp.Pool(processes=mp.cpu_count()//2)
    instantiatePool(pool)

    logging.info("Loading projects.")

    # Reset projects to empty list
    loadedProjects = []
    notProcessedFiles = []

    # Load projects from the database
    projs = models.Project.query.all()
    for p in projs:

        logging.info(f"Loading project id {p.id} ({p.name}) at {p.path}")

        # Path object for the project
        projDirPathObj = Path(p.path)

        # Validate the project folder
        try:
            validateProjectFolder(projDirPathObj)
        except Exception as e:
            logging.error(f'Project folder {projDirPathObj} is invalid.\n{e}\n{traceback.format_exc()}')
            # TODO(gus): Update the database to reflect this?
        else:
            # Instantiate project, and add to the list to be returned
            # TODO(gus): We need to have project take absolute path and project name!
            loadedProjects.append(Project(p))

            logging.info("Finished loading project.")

    # Detect new project folders not in the database
    for projDirPathObj in [p for p in config['projectsDirPathObj'].iterdir() if p.is_dir()]:

        # Check whether path exists in database projects
        foundInExistingProject = False
        for p in projs:
            if Path(p.path).samefile(projDirPathObj):
                foundInExistingProject = True
                break

        # Skip if this project folder was found to exist in the database already
        if foundInExistingProject:
            continue

        logging.info(f"Found new project at {projDirPathObj}.")

        # Ensure we have a valid project folder (i.e. it can be empty, but it
        # cannot contain invalid files or folders).
        try:
            validateProjectFolder(projDirPathObj)
        except Exception as e:
            logging.error(f"Project folder {projDirPathObj} is invalid.\n{e}\n{traceback.format_exc()}")
        else:
            logging.info(f'Project folder {projDirPathObj} is valid.')

            # Scaffold project folder
            scaffoldProjectFolder(projDirPathObj)

            # Add project to the database
            project = models.Project(name=projDirPathObj.name, path=str(projDirPathObj.resolve()))
            models.db.session.add(project)
            models.db.session.commit()
            logging.info(f"Added project to database: {projDirPathObj}")

            # Instantiate project
            # TODO(gus): We need to have project take absolute path and project name!
            loadedProjects.append(Project(project))

    # Delete all files that may have been downsampled incorrectly and adds unprocessed files to be downsampled
    for project in loadedProjects:
        for projFile in project.files:
            tmp_file = Path(str(projFile.procFilePathObj)+'.tmp')
            if tmp_file.exists():
                try:
                    projFile.procFilePathObj.unlink(missing_ok=True)
                except Exception as e:
                    raise RuntimeError(f"Downsample file {str(projFile.procFilePathObj)} exists and could not be deleted. The viewer will quit now. It is recommended that the processed file be deleted manually. \n{e}\n{traceback.format_exc()}")
                else:
                    tmp_file.unlink()

                logging.info(f"Partial processed file {str(projFile.procFilePathObj)} has been removed.")
            
            # Add all non processed files for downsampling
            if not projFile.procFilePathObj.exists():
                notProcessedFiles.append((str(projFile.origFilePathObj.resolve()), str(projFile.procFilePathObj.parent.resolve())))

    for downsampParam in notProcessedFiles:
        downsamplePool.apply_async(downsampleFile, downsampParam)


    logging.info("Finished loading projects.")

    return getProjects()

def scaffoldProjectFolder(projDirPathObj):
    """Generate the baseline project folder contents as needed"""

    logging.info(f"Scaffolding project folder {projDirPathObj}.")

    # Create the originals folder if needed
    (projDirPathObj / 'originals').mkdir(exist_ok=True)

    # Create the processed folder if needed
    (projDirPathObj / 'processed').mkdir(exist_ok=True)

    # Create the templates folder if needed
    projTemplatesFolderPathObj = projDirPathObj / 'templates'
    projTemplatesFolderPathObj.mkdir(exist_ok=True)

    # Create empty template files if needed
    createEmptyJSONFile(projTemplatesFolderPathObj / 'interface_templates.json')
    createEmptyJSONFile(projTemplatesFolderPathObj / 'project_template.json')

    logging.info(f"Finished scaffolding project folder {projDirPathObj}.")

def setDataPath(path, load_projects=False) -> None:
    """
    Loads a data path for use via Python (instead of as a web server). If the
    optional load_projects parameter is False, then projects must be loaded
    manually using loadProject() or loadProjects().
    """

    set_data_path(path)

    app = Flask(__name__)
    app.config.update({
        'SQLALCHEMY_TRACK_MODIFICATIONS': False,
        'SQLALCHEMY_DATABASE_URI': f"sqlite:///{(config['dataPathObj'] / 'database' / 'db.sqlite')}",
    })

    models.init_flask_app(app)
    app.app_context().push()

    if load_projects:
        loadProjects()

def validateProjectFolder(projDirPathObj):
    """Raises an exception if the project folder is invalid"""

    logging.info(f"Validating project folder {projDirPathObj}.")

    # Validate top-level folders, excluding dot-files
    for p in [p for p in projDirPathObj.iterdir() if not p.name.startswith('.')]:

        # Project folder should only contain directories
        if not p.is_dir():
            raise Exception(f'Project folder contains invalid file: {p}')

        # Validate expected top-level folder names
        if p.name not in ['originals', 'processed', 'templates']:
            raise Exception(f'Project folder contains invalid folder: {p}')

    # Validate expected template files if they exist.
    tp = projDirPathObj / "templates"
    if tp.exists():
        for p in [p for p in tp.iterdir()]:
            if p.name not in ['project_template.json', 'interface_templates.json'] and not p.name.startswith('.'):
                raise Exception(f'Project templates folder contains invalid file: {p}')

    # TODO(gus): Validate that originals have processed?

    logging.info(f'Finished validating project folder {projDirPathObj}.')



""" This file contains functions to generate and verify tokens for Flask-User.
    Tokens contain an encoded user ID and a signature. The signature is managed by the itsdangerous module.

    :copyright: (c) 2013 by Ling Thio
    :author: Ling Thio (ling.thio@gmail.com)
    :license: Simplified BSD License, see LICENSE.txt for more details."""

import base64
from Crypto.Cipher import AES
from itsdangerous import BadSignature, SignatureExpired, TimestampSigner

class TokenManager(object):
    def setup(self, secret):
        """ Create a cypher to encrypt IDs and a signer to sign tokens."""
        # Create cypher to encrypt IDs
        # and ensure >=16 characters
        precursor = b'0123456789abcdef'
        if isinstance(secret, bytes):
            key = secret + precursor
        else:
            key = secret.encode("utf-8") + precursor
        self.cipher = AES.new(key[0:16], AES.MODE_ECB)

        # Create signer to sign tokens
        self.signer = TimestampSigner(secret)

    def encrypt_id(self, id):
        """ Encrypts integer ID to url-safe base64 string."""
        # 16 byte integer
        str1 = '%016d' % id
        # encrypted data
        str2 = self.cipher.encrypt(str1.encode())
        # URL safe base64 string with '=='
        str3 = base64.urlsafe_b64encode(str2)
        # return base64 string without '=='
        return str3[0:-2]

    def decrypt_id(self, encrypted_id):
        """ Decrypts url-safe base64 string to integer ID"""
        # Convert strings and unicode strings to bytes if needed
        if hasattr(encrypted_id, 'encode'):
            encrypted_id = encrypted_id.encode('ascii', 'ignore')

        try:
            str3 = encrypted_id + b'=='             # --> base64 string with '=='
            #print('str3=', str3)
            str2 = base64.urlsafe_b64decode(str3)   # --> encrypted data
            #print('str2=', str2)
            str1 = self.cipher.decrypt(str2)        # --> 16 byte integer string
            #print('str1=', str1)
            return int(str1)                        # --> integer id
        except Exception as e:                      # pragma: no cover
            print('!!!Exception in decrypt_id!!!:', e)
            return 0

    def generate_token(self, id):
        """ Return token with id, timestamp and signature"""
        # In Python3 we must make sure that bytes are converted to strings.
        # Hence the addition of '.decode()'
        return self.signer.sign(self.encrypt_id(id)).decode()

    def verify_token(self, token, expiration_in_seconds):
        """ Verify token and return (is_valid, has_expired, id).
            Returns (True, False, id) on success.
            Returns (False, True, None) on expired tokens.
            Returns (False, False, None) on invalid tokens."""
        try:
            data = self.signer.unsign(token, max_age=expiration_in_seconds)
            is_valid = True
            has_expired = False
            id = self.decrypt_id(data)
        except SignatureExpired:
            is_valid = False
            has_expired = True
            id = None
        except BadSignature:
            is_valid = False
            has_expired = False
            id = None
        return (is_valid, has_expired, id)



"""Class and related functionality for pattern sets."""

from sqlalchemy.orm import joinedload
import pandas as pd
from typing import Union, List

from . import models
from .shared import annotationDataFrame, patternDataFrame

class PatternSet:
    """Represents a pattern set."""

    def __init__(self, projparent, dbmodel):

        # Holds reference to the project parent
        self.projparent = projparent

        # Set id & name
        self.id = dbmodel.id
        self.name = dbmodel.name
        self.showByDefault = dbmodel.show_by_default

        # Holds the db model
        self.dbmodel = dbmodel

        # Establish a count of patterns belonging to the set
        self.count = 0

        # Refresh model & count
        self.refresh()

    def addPatterns(self, df, validate=True):
        """
        Add patterns to the pattern set. By default, the rows will be validated (e.g. for matching file ID & filename).
        This may be skipped in the case of extremely high volume, but it may lead to database integrity issues to do so.

        During validation, if filename is present and file_id is not, then file_id will be populated according to the
        filename. If both are populated, then the file_id will be validated to match the filename. The provided pattern
        set must contain 'file_id' and/or 'filename' columns as well as ['series', 'left', 'right', 'label'].
        :return: None
        """

        if validate:

            if 'file_id' not in df.columns and 'filename' not in df.columns:
                raise Exception('Pattern set must have either a file_id or filename column. Neither was found.')

            required_columns = ['series', 'left', 'right', 'label']
            if not all(c in df.columns for c in required_columns):
                raise Exception(f"Pattern set is missing one of the following required columns: [{', '.join(required_columns)}]")

            # Add optional columns if not present
            if 'file_id' not in df.columns:
                df['file_id'] = None
            if 'top' not in df.columns:
                df['top'] = None
            if 'bottom' not in df.columns:
                df['bottom'] = None

            # Reset index (so row number corresponds to index)
            df.reset_index(drop=True, inplace=True)

            # We'll do a validation and/or file_id fill-in if the filename column is present
            checkfn = 'filename' in df.columns

            for i in range(df.shape[0]):
                row = df.loc[i]
                fid = row['file_id']
                series = row['series']
                left = row['left']
                right = row['right']

                # These columns might be NaN (which evaluates to truthy), so override them to None
                if pd.isnull(fid):
                    fid = None
                if pd.isnull(left):
                    left = None
                if pd.isnull(right):
                    right = None

                # Check filename
                if checkfn and row['filename']:
                    fn = row['filename']
                    f = self.projparent.getFileByName(fn)
                    if not f:
                        # If they provided a filename but the file wasn't found in the project, raise an exception.
                        raise Exception(f"File {fn} not found in the project.")
                    if fid:
                        # If they provided both file_id and filename but they don't match, raise an exception.
                        if f.id != fid:
                            raise Exception(f"File ID {f.id} for filename {fn} did not match the provided file_id {fid}.")
                    else:
                        # File ID not provided, so populate it.
                        df.at[i, 'file_id'] = f.id
                    # Note: If the file_id was provided and filename was not provided, we don't care, as the
                    # filename column will be dropped.

                # If file_id is still not populated for any reason, we have a problem
                fid = df.at[i, 'file_id']
                if not fid or pd.isnull(fid):
                    raise Exception(f"File ID not found for {row}.")

                # Validate the file ID
                f = self.projparent.getFile(fid)
                if not f:
                    raise Exception(f"File ID {fid} not found for {row}.")

                # Check series name
                # TODO(gus, vedant): Check series name, but the file loading issue has to be resolved first.
                if not series:# or not f.getSeries(series):
                    raise Exception(f"Series {series} not found for {row}.")

                # Check left & right
                if not left or not right:
                    raise Exception(f"Left or right not found for {row}.")

        # Subset only the columns we need from the user
        df = df[['file_id', 'series', 'left', 'right', 'top', 'bottom', 'label']]

        # Add the project ID
        df.insert(0, "project_id", [self.projparent.id]*df.shape[0])

        # Add the id of this pattern set
        df.insert(0, "pattern_set_id", [self.id]*df.shape[0])

        # Do the db insert
        df.to_sql('patterns', models.db.engine, index=False, if_exists='append')

        # Refresh & update count
        self.refresh()

    def assignToUsers(self, user_ids: Union[int, List[int]]) -> None:
        """
        Assign the pattern set to user(s). Idempotent.
        :param user_ids: May be single user ID or list of user IDs.
        :return: None
        """
        if not isinstance(user_ids, list):
            user_ids = [user_ids]
        self.dbmodel.users.extend(
            models.User.query.filter(models.User.id.in_(user_ids)).all()
        )
        models.db.session.commit()

    def delete(self, deletePatterns=False):
        """
        Deletes the pattern set from the database and the parent project
        instance. If the pattern set has patterns, the deletion will fail,
        unless the deletePatterns flag is True, in which case it will first
        delete the child patterns.
        """
        if deletePatterns:
            self.deletePatterns()
        models.db.session.rollback()
        try:
            models.db.session.delete(self.dbmodel)
        except:
            models.db.session.rollback()
            raise
        models.db.session.commit()
        del self.projparent.patternsets[self.id]

    def deletePatterns(self) -> int:
        """
        Delete the patterns belonging to this pattern set.
        :return: number of deleted patterns
        """
        models.db.session.rollback()
        try:
            n = models.Pattern.query.filter_by(pattern_set_id=self.id).delete()
        except:
            models.db.session.rollback()
            raise
        models.db.session.commit()
        self.refresh()
        return n

    def deleteUnannotatedPatterns(self) -> int:
        """
        Delete all patterns which have not yet been annotated from the set.
        :return: number of deleted patterns
        """
        models.db.session.rollback()
        try:
            n = models.Pattern.query.filter(models.Pattern.pattern_set_id == self.id, models.Pattern.id.notin_(
                models.db.session.query(models.Annotation.pattern_id).filter(models.Annotation.pattern_id.isnot(None)).subquery()
            )).delete(synchronize_session=False)
        except:
            models.db.session.rollback()
            raise
        models.db.session.commit()
        self.refresh()
        return n

    def getAnnotationCount(self) -> int:
        """Returns a count of annotations which annotate any pattern in this set."""
        return models.Annotation.query.filter_by(pattern_set_id=self.id).count()

    def getAnnotations(self) -> pd.DataFrame:
        """Returns a DataFrame of the annotations in this set."""
        return annotationDataFrame(models.Annotation.query.options(joinedload('user')).filter_by(pattern_set_id=self.id).all())

    def getPatternCount(self) -> int:
        """Returns a count of the patterns in this set."""
        return self.count

    def getPatterns(self) -> pd.DataFrame:
        """Returns a DataFrame of the patterns in this set."""
        return patternDataFrame(self.dbmodel.patterns)

    def setDescription(self, description: str):
        """Set the pattern set's description."""
        self.dbmodel.description = description
        models.db.session.commit()

    def setName(self, name: str):
        """Set the pattern set's name."""
        self.dbmodel.name = name
        models.db.session.commit()
        self.name = name

    def setShowByDefault(self, show: bool):
        """Set whether a pattern set should show by default."""
        self.dbmodel.show_by_default = show
        models.db.session.commit()
        self.showByDefault = show

    def refresh(self):
        """
        Refresh model & update the count of patterns belonging to this set
        (this is normally an internally-used method).
        """
        models.db.session.refresh(self.dbmodel)
        self.count = models.Pattern.query.filter_by(pattern_set_id=self.id).count()

def getAssignmentsPayload(user_id):
    return [{
        'id': ps.id,
        'name': ps.name,
        'description': ps.description,
        'project_id': ps.project.id,
        'project_name': ps.project.name,
        'completed': models.Annotation.query.filter_by(user_id=user_id, pattern_set_id=ps.id).count(),
        'remaining': len(ps.patterns) - models.Annotation.query.filter_by(user_id=user_id, pattern_set_id=ps.id).count(),
        'total': len(ps.patterns),
    } for ps in models.PatternSet.query.filter(models.PatternSet.users.any(id=user_id)).all()]


"""
    tests.test_all_forms.py
    ---------------------------
    Flask-User automated tests:
    Tests posting a sequence of valid forms for all possible config combinations

    :copyright: (c) 2013 by Ling Thio
    :author: Ling Thio (ling.thio@gmail.com)
    :license: Simplified BSD License, see LICENSE.txt for more details.
"""

from __future__ import print_function

import datetime

from flask import current_app, url_for

# **********************
# ** Global Variables **
# **********************
VALID_USERNAME = 'valid1'
VALID_EMAIL = 'valid1@example.com'
VALID_PASSWORD = 'Password1'
INVITE_USER_EMAIL = 'valid2@example.com'
# Using global variable for speed
valid_user = None
valid_user_invite = None

# ********************************
# ** Automatically called Tests **
# ********************************
# The 'client' and 'app' parameters are set up in conftest.py
# Functions that start with 'test' will be run automatically by the test suite runner (py.test)

def test_with_email(client):
    """
    Test all forms with all enabled features and enable_username=False
    """
    print('test_with_email()')

    um = current_app.user_manager
    um.enable_register = True
    um.enable_username = False
    um.enable_email = True          # Login with email
    um.enable_confirm_email = True
    um.enable_change_username = False
    um.enable_change_password = True
    um.enable_forgot_password = True
    um.enable_multiple_emails = False
    um.enable_retype_password = True

    check_all_valid_forms(um, client)

def test_with_username(client):
    """
    Test all forms with all enabled features and enable_username=True
    """
    print('test_with_username()')

    # Enable all features
    um = current_app.user_manager
    um.enable_register = True
    um.enable_username = True       # Login with username
    um.enable_email = False
    um.enable_confirm_email = False
    um.enable_change_username = True
    um.enable_change_password = True
    um.enable_forgot_password = False
    um.enable_multiple_emails = False
    um.enable_retype_password = True

    check_all_valid_forms(um, client)

# *****************************
# ** Explicitly called Tests **
# *****************************
# These function names may not start with 'test'

def do_test_all_possible_config_combinations(client, db):
    """
    Test all forms with all possible config combinations
    """
    print()
    print("Testing all forms for all possible config combinations")
    um =  current_app.user_manager

    for um.enable_register in (True, False):
      for um.enable_email in (True, False):
        for um.enable_retype_password in (True, False):
          for um.enable_confirm_email in (True, False):
            for um.enable_username in (True, False):
              for um.enable_change_password in (True, False):
                for um.enable_change_username in (True, False):
                  for um.enable_forgot_password in (True, False):
                    for um.enable_invitation in (True, False):
                        check_all_valid_forms(um, client, db)

# **************************
# ** Check Form Functions **
# **************************
# Below we check each form by submitting fields depending on the config settings.

def check_all_valid_forms(um, client):
    # ** Skip tests for invalid config combinations
    # USER_ENABLE_REGISTER=True must have USER_ENABLE_USERNAME=True or USER_ENABLE_EMAIL=True or both.
    if um.enable_register and not um.enable_email and not um.enable_username: return
    # USER_ENABLE_CONFIRM_EMAIL=True must have USER_ENABLE_EMAIL=True
    if um.enable_confirm_email and not um.enable_email: return
    # USER_ENABLE_MULTIPLE_EMAILS=True must have USER_ENABLE_EMAIL=True
    if um.enable_multiple_emails and not um.enable_email: return
    # ENABLE_CHANGE_USERNAME=True must have ENABLE_USERNAME=True.
    if um.enable_change_username and not um.enable_username: return

    check_valid_register_form(um, client, client.db)
    check_valid_resend_confirm_email_form(um, client)
    check_valid_confirm_email_page(um, client)
    check_valid_login_form(um, client)
    check_valid_change_password_form(um, client)
    check_valid_change_username_form(um, client)
    check_valid_logout_link(um, client)
    check_valid_forgot_password_form(um, client)
    check_valid_reset_password_page(um, client)
    check_valid_invite_email(um, client)
    #check_valid_invite_registration_different_email(um, client)

    delete_valid_user(client.db)
    delete_valid_user_invite(client.db)

def check_valid_register_form(um, client, db):
    # Using global variable for speed
    global valid_user
    User = um.db_adapter.UserClass

    # Build variable argument list depending on config settings
    kwargs = {}
    if um.enable_username:
        kwargs['username'] = VALID_USERNAME
    if um.enable_email:
        kwargs['email'] = VALID_EMAIL
    kwargs['password'] = VALID_PASSWORD

    if um.enable_register:
        print("test_valid_register_form")

        # Create User by submitting a form
        if um.enable_retype_password:
            kwargs['retype_password'] = VALID_PASSWORD

        # Submit form and verify that response has no errors
        client.post_valid_form(url_for('user.register'), **kwargs)

        if um.enable_username:
            valid_user = User.query.filter(User.username==VALID_USERNAME).first()
        else:
            valid_user = User.query.filter(User.email==VALID_EMAIL).first()
        assert valid_user

        # Verify operations
        assert valid_user.active

    else:
        # Create user record manually

        # hash password
        kwargs['password'] = um.hash_password(VALID_PASSWORD)

        # Create User
        valid_user = User(active=True, confirmed_at=datetime.datetime.utcnow(), **kwargs)
        db.session.add(valid_user)
        db.session.commit()
        assert valid_user

def check_valid_resend_confirm_email_form(um, client):
    # Skip test for certain config combinations
    if not um.enable_register: return
    if not um.enable_email: return
    if not um.enable_confirm_email: return

    print("test_valid_resend_confirm_email_form")

    # Submit form and verify that response has no errors
    client.post_valid_form(url_for('user.resend_confirm_email'), email=VALID_EMAIL)

def check_valid_confirm_email_page(um, client):
    # Skip test for certain config combinations
    if not um.enable_register: return
    if not um.enable_email: return
    if not um.enable_confirm_email: return

    print("test_valid_confirm_email_page")

    # Generate confirmation token for user 1
    confirmation_token = um.generate_token(valid_user.id)

    # Retrieve page and verify that response has no errors
    client.get_valid_page(url_for('user.confirm_email', token=confirmation_token))

    # Verify operations
    assert valid_user.active
    assert valid_user.confirmed_at != None

def check_valid_login_form(um, client):
    print("test_valid_login_form")

    # Build variable argument list depending on config settings
    kwargs = {}
    if um.enable_username:
        kwargs['username'] = VALID_USERNAME
    if um.enable_email:
        kwargs['email'] = VALID_EMAIL
    kwargs['password'] = VALID_PASSWORD

    # Submit form and verify that response has no errors
    client.post_valid_form(url_for('user.login'), **kwargs)

    # Verify operations
    # TODO:

def check_valid_change_password_form(um, client):
    # Skip test for certain config combinations
    if not um.enable_change_password: return

    print("test_valid_change_password_form")

    # Define defaults
    new_password = 'Password9'
    old_hashed_password = valid_user.password

    # Build variable argument list depending on config settings
    kwargs = {}
    kwargs['old_password'] = VALID_PASSWORD
    kwargs['new_password'] = new_password
    if um.enable_retype_password:
        kwargs['retype_password'] = new_password

    # Submit form and verify that response has no errors
    client.post_valid_form(url_for('user.change_password'), **kwargs)

    # Verify operations
    assert um.verify_password(new_password, valid_user)

    # Change password back to old password for subsequent tests
    valid_user.password = old_hashed_password

def check_valid_change_username_form(um, client):
    # Skip test for certain config combinations
    if not um.enable_change_username: return

    print("test_valid_change_username_form")

    new_username = 'user9'

    # Submit form and verify that response has no errors
    client.post_valid_form(url_for('user.change_username'), new_username=new_username, old_password=VALID_PASSWORD)

    # Verify operations
    assert valid_user.username == new_username

    # Change username back to old password for subsequent tests
    valid_user.username = VALID_USERNAME

def check_valid_logout_link(um, client):
    print("test_valid_logout_link")
    # Retrieve page and verify that response has no errors
    client.get_valid_page(url_for('user.logout'))

def check_valid_forgot_password_form(um, client):
    # Skip test for certain config combinations
    if not um.enable_email: return
    if not um.enable_forgot_password: return

    print("test_valid_forgot_password_form")

    # Submit form and verify that response has no errors
    client.post_valid_form(url_for('user.forgot_password'), email=VALID_EMAIL)

def check_valid_reset_password_page(um, client):
    # Skip test for certain config combinations
    if not um.enable_email: return
    if not um.enable_forgot_password: return

    print("test_valid_reset_password_page")

    # Simulate a valid forgot password form
    token = um.generate_token(valid_user.id)

    # Define defaults
    password = 'Password1'
    new_password = 'Password9'
    old_hashed_password = valid_user.password
    url = url_for('user.reset_password', token=token)

    # Build variable argument list depending on config settings
    kwargs = {}
    kwargs['new_password'] = new_password
    if um.enable_retype_password:
        kwargs['retype_password'] = new_password

    # Submit form and verify that response has no errors
    client.post_valid_form(url, **kwargs)

    # Verify operations
    assert um.verify_password(new_password, valid_user)

    # Change password back to old password for subsequent tests
    valid_user.password = old_hashed_password

def check_valid_invite_email(um, client):
    """ If a valid email is submitted using the invite form,
    then it should generate the proper email and response """
    if not um.enable_invitation: return
    # Submit form and verify that response has no errors
    global valid_user_invite
    UserInvite = um.db_adapter.UserInvitationClass
    client.login(username='member', email='member@example.com', password='Password1')
    client.post_valid_form(url_for('user.invite'), email=INVITE_USER_EMAIL)
    valid_user_invite = UserInvite.query.filter(UserInvite.email==INVITE_USER_EMAIL).first()
    assert valid_user_invite

def delete_valid_user(db):
    # Using global variable for speed
    global valid_user

    # Delete valid_user
    db.session.delete(valid_user)
    db.session.commit()
    valid_user = None

def delete_valid_user_invite(db):
    global valid_user_invite

    db.session.delete(valid_user_invite)
    db.session.commit()
    valid_user_invite = None



from .model import SimpleFeaturizer, FeaturizerParameter

class CoeffOfVariationFeaturizer(SimpleFeaturizer):

    id = 'cv'
    name = 'Coefficient of Variation'
    description = 'Coefficient of variation is the mean over the standard deviation.'
    parameters = [
        FeaturizerParameter(id='skipna', name="Skip NaN", description="Skip NaN values", data_type='boolean'),
    ]

    def featurize(self, data, params={}):
        skipna = params['skipna']
        return data.mean(skipna=skipna)/data.std(skipna=skipna)

class MADFeaturizer(SimpleFeaturizer):

    id = 'mad'
    name = 'Median of Absolute Deviation from Mean'

    def featurize(self, data, params={}):
        # TODO(gus): Add skipna?
        return (data - data.mean()).abs().median()

class NFeaturizer(SimpleFeaturizer):

    id = 'n'
    name = 'Value Count (n)'

    def featurize(self, data, params={}):
        # TODO(gus): Add ability to skipna
        return data.shape[0]

class MinFeaturizer(SimpleFeaturizer):

    id = 'min'
    name = 'Min'
    parameters = [
        FeaturizerParameter(id='skipna', name="Skip NaN", description="Skip NaN values", data_type='boolean'),
    ]

    def featurize(self, data, params={}):
        skipna = params['skipna']
        return data.min(skipna=skipna)

class MaxFeaturizer(SimpleFeaturizer):

    id = 'max'
    name = 'Max'
    parameters = [
        FeaturizerParameter(id='skipna', name="Skip NaN", description="Skip NaN values", data_type='boolean'),
    ]

    def featurize(self, data, params={}):
        skipna = params['skipna']
        return data.max(skipna=skipna)

class MedianFeaturizer(SimpleFeaturizer):

    id = 'median'
    name = 'Median'
    parameters = [
        FeaturizerParameter(id='skipna', name="Skip NaN", description="Skip NaN values", data_type='boolean'),
    ]

    def featurize(self, data, params={}):
        skipna = params['skipna']
        return data.median(skipna=skipna)

class RangeFeaturizer(SimpleFeaturizer):

    id = 'range'
    name = 'Range'
    description = 'Range is max minus min.'
    parameters = [
        FeaturizerParameter(id='skipna', name="Skip NaN", description="Skip NaN values", data_type='boolean'),
    ]

    def featurize(self, data, params={}):
        skipna = params['skipna']
        return data.max(skipna=skipna) - data.min(skipna=skipna)

class RangeRatioFeaturizer(SimpleFeaturizer):

    id = 'rangeratio'
    name = 'Range Ratio (range / median)'
    description = 'Range ratio is the range over median.'
    parameters = [
        FeaturizerParameter(id='skipna', name="Skip NaN", description="Skip NaN values", data_type='boolean'),
    ]

    def featurize(self, data, params={}):
        skipna = params['skipna']
        return (data.max(skipna=skipna) - data.min(skipna=skipna)) / data.median(skipna=skipna)

class DataDenFeaturizer(SimpleFeaturizer):

    id = 'dataden'
    name = 'Data Density'
    description = 'Data density is the number of values (n) over the window size in seconds'

    def featurize(self, data, params={}):
        # TODO(gus): Add ability to skipna
        # TODO(gus): Move this exception handling to outer base class
        try:
            return data.shape[0] / (data.index[-1].timestamp() - data.index[0].timestamp())
        except:
            return None

from sklearn.linear_model import LinearRegression
slrm = LinearRegression()
class LRSlopeFeaturizer(SimpleFeaturizer):

    id = 'skl_slope'
    name = 'Linear Regression Slope (sklearn)'
    description = 'Slope of linear regression computed with sklearn.LinearRegression'

    def featurize(self, data, params={}):
        # TODO(gus): Move this exception handling to outer base class
        try:
            slrm.fit(data.reset_index()[['time']], data)
            return slrm.coef_[0]
        except:
            return None



### BELOW HERE NEEDS TESTING


class MaxGapFeaturizer(SimpleFeaturizer):

    id = 'maxgap'
    name = 'Max Gap'
    description = 'Maximum timestamp gap between two subsequent measurements'

    def featurize(self, data, params={}):
        # TODO(gus): Move this exception handling to outer base class
        try:
            return data.index.diff().max()
        except:
            return None



import rpy2
import rpy2.robjects as robjects
from rpy2.robjects import pandas2ri
from rpy2.robjects.conversion import localconverter
from rpy2.robjects.packages import importr
robjects.r['source']('temp.R')

import numpy as np
class RobustSlopeFeaturizer(SimpleFeaturizer):

    id = 'robustslope'
    name = 'Robst Slope (R)'
    description = 'Slope from the robust linear regression fit using M estimator (uses R)'

    def featurize(self, data, params={}):
        with localconverter(robjects.default_converter + pandas2ri.converter):
            # ret = robjects.r['get_robust_slope'](data.index, data['value'])
            # ret = robjects.r['get_robust_slope'](data.reset_index()['time'].values.astype(np.int64) // 10 ** 9, data)

            data = data.reset_index() # data.reset_index(inplace=True)
            data['time'] = data['time'].astype(np.int64) // 10**9



            # t = data.reset_index()[['time']]
            # t['time'] = t['time'].astype(np.int64) // 10**9
            # data['time'] = t['time']
            # print(t)
            print('arg1')
            print(data['time'])
            print('arg2')
            print(data['value'])
            # ret = robjects.r['get_robust_slope'](data[['time']], data.set_index('time'))
            ret = robjects.r['get_robust_slope'](data['time'], data['value'])
        print(ret)
        print(ret[0])
        return None if isinstance(ret[0], rpy2.rinterface_lib.sexp.NALogicalType) else ret[0]


"""Class and related functionality for projects."""
import importlib
import numpy as np

import datetime as dt
from io import BytesIO
import logging
import math
import pandas as pd
import pickle
import traceback
import random
from collections import Counter

from pathlib import Path
from sqlalchemy import and_
from sqlalchemy.orm import contains_eager, joinedload
from typing import AnyStr, List, Dict, Optional, Union

from . import models
from .patternset import PatternSet
from .config import config
from .file import File
from .shared import annotationDataFrame, annotationOrPatternOutput, getProcFNFromOrigFN, patternDataFrame

class Project:
    """Represents an auviewer project."""

    def __init__(self, projectModel, processNewFiles=True):
        """The project name should also be the directory name in the projects directory."""
        
        # Set id & name
        self.id = projectModel.id
        self.name = projectModel.name

        print(f"Loading project ID {self.id} ({self.name})... ", end='')
        
        # Set relevant paths
        self.projDirPathObj = Path(projectModel.path)
        self.originalsDirPathObj = self.projDirPathObj / 'originals'
        self.processedDirPathObj = self.projDirPathObj / 'processed'

        # Hold a reference to the db model for future use
        self.model = projectModel

        # Load interface templates
        self.interfaceTemplates = "{}"
        p = self.projDirPathObj / 'templates' / 'interface_templates.json'
        if p.is_file():
            with p.open() as f:
                self.interfaceTemplates = f.read()

        # Load project template
        self.projectTemplate = "{}"
        p = self.projDirPathObj / 'templates' / 'project_template.json'
        if p.is_file():
            with p.open() as f:
                self.projectTemplate = f.read()

        # Holds references to the files that belong to the project
        self.files = []

        # Holds references to the pattern sets that belong to the project,
        # indexed by pattern set ID.
        self.patternsets = {}

        # Load pattern sets
        self.loadPatternSets()

        # Load project files
        self.loadProjectFiles(processNewFiles)

        print(f"Complete")

    def createPatternSet(self, name: str, description=None, showByDefault: bool = True) -> PatternSet:
        """
        Create and return a new pattern set.
        :return: a new PatternSet instance
        """

        # Create pattern set in the database
        patternSetModel = models.PatternSet(project_id=self.id, name=name, description=description, show_by_default=showByDefault)
        models.db.session.add(patternSetModel)
        models.db.session.commit()

        # Instantiate PatternSet and add to the project's pattern sets
        ps = PatternSet(self, patternSetModel)
        self.patternsets[ps.id] = ps

        # Return the pattern set
        return ps

    def detectPatterns(
            self, 
            type, 
            series, 
            thresholdlow, 
            thresholdhigh, 
            duration, 
            persistence, 
            maxgap, 
            expected_frequency=0, 
            min_density=0,
            drop_values_below=None,
            drop_values_above=None,
            drop_values_between=None,
        ):
        """
        Run pattern detection on all files, and return a DataFrame of results.
        This DataFrame, or a subset thereof, can be passed into PatternSet.addPatterns() if desired.
        """
        patterns = [[f.id, f.name, series, pattern[0], pattern[1], None, None] for f in self.files for pattern in f.detectPatterns(
            type, 
            series, 
            thresholdlow, 
            thresholdhigh, 
            duration, 
            persistence, 
            maxgap, 
            expected_frequency=expected_frequency, 
            min_density=min_density, 
            drop_values_below=drop_values_below, 
            drop_values_above=drop_values_above, 
            drop_values_between=drop_values_between
        )]
        pdf = pd.DataFrame(patterns, columns=['file_id', 'filename', 'series', 'left', 'right', 'top', 'bottom'])
        pdf['label'] = ''
        return pdf

    def getAnnotations(
            self,
            annotation_id: Union[int, List[int], None] = None,
            file_id: Union[int, List[int], None] = None,
            pattern_id: Union[int, List[int], None] = None,
            pattern_set_id: Union[int, List[int], None]=None,
            series: Union[AnyStr, List[AnyStr], None]=None,
            user_id: Union[int, List[int], None] = None) -> pd.DataFrame:
        """
        Returns a dataframe of annotations for this project, optionally filtered.
        """

        # Prepare input
        if not isinstance(annotation_id, List) and annotation_id is not None:
            annotation_id = [annotation_id]
        if not isinstance(file_id, List) and file_id is not None:
            file_id = [file_id]
        if not isinstance(pattern_id, List) and pattern_id is not None:
            pattern_id = [pattern_id]
        if not isinstance(pattern_set_id, List) and pattern_set_id is not None:
            pattern_set_id = [pattern_set_id]
        if not isinstance(series, List) and series is not None:
            series = [series]
        if not isinstance(user_id, List) and user_id is not None:
            user_id = [user_id]

        # Query
        q = models.Annotation.query.options(joinedload(models.Annotation.user))

        # Filter query as necessary
        if annotation_id is not None:
            q = q.filter(models.Annotation.id.in_(annotation_id))
        if file_id is not None:
            q = q.filter(models.Annotation.file_id.in_(file_id))
        if pattern_id is not None:
            q = q.filter(models.Annotation.pattern_id.in_(pattern_id))
        if pattern_set_id is not None:
            q = q.filter(models.Annotation.pattern_set_id.in_(pattern_set_id))
        if series is not None:
            q = q.filter(models.Annotation.series.in_(series))
        if user_id is not None:
            q = q.filter(models.Annotation.user_id.in_(user_id))

        # Return the dataframe
        return annotationDataFrame(q.filter(models.Annotation.project_id == self.id).all())

    def getAnnotationsOutput(self, user_id: int):
        """Returns a list of user's annotations for all files in the project"""
        return [[a.id, a.file_id, Path(a.file.path).name, a.series, a.left, a.right, a.top, a.bottom, a.label, a.pattern_id] for a in models.Annotation.query.filter_by(user_id=user_id, project_id=self.id).all()]

    def getFile(self, id):
        """Returns the file with matching ID or None."""
        for f in self.files:
            if f.id == id:
                return f
        return None

    def getFileByName(self, name):
        """Returns the file with matching ID or None."""
        for f in self.files:
            if f.name == name:
                return f
        return None

    def makeFilesPayload(self, files):
        # for f in files:
        #     f.initfile()
        outputObject = {
            'files': [[f.id, f.origFilePathObj.name] for f in files],
            'series': [],
            'events': [],#[f.getEvents() for f in self.files],
            'metadata': [f.getMetadata() for f in files]
        }

        #must populate outputObject with constituent files' series, events, and metadata
        for f in files:
            s = self.getSeriesToRender(f)
            if (s):
                outputObject['series'].append({s.id:  s.getFullOutput()})
            else:
                outputObject['series'].append({None: None})

        return outputObject

    def getConstituentFilesPayload(self):
        files = self.files

        return self.makeFilesPayload(files)

    def getInitialPayload(self, user_id):
        """Returns initial project payload data"""

        print("Assembling initial project payload output for project", self.name)

        return {

            # Project data
            'project_id': self.id,
            'project_name': self.name,
            'project_assignments': [{
                'id': ps.id,
                'name': ps.name,
                'description': ps.description,
                'patterns': [
                    annotationOrPatternOutput(p, p.annotations[0] if len(p.annotations) > 0 else None) for p in models.db.session.query(models.Pattern).filter_by(pattern_set_id=ps.id).outerjoin(models.Annotation, and_(models.Annotation.pattern_id == models.Pattern.id, models.Annotation.user_id == user_id)).options(contains_eager(models.Pattern.annotations)).all()
                ]
            } for ps in models.PatternSet.query.filter(models.PatternSet.users.any(id=user_id), models.PatternSet.project_id==self.id).all()],
            'project_files': [[f.id, f.origFilePathObj.name] for f in self.files],

            # Template data
            'builtin_default_interface_templates': config['builtinDefaultInterfaceTemplates'],
            'builtin_default_project_template': config['builtinDefaultProjectTemplate'],
            'global_default_interface_templates': config['globalDefaultInterfaceTemplates'],
            'global_default_project_template': config['globalDefaultProjectTemplate'],
            'interface_templates': self.interfaceTemplates,
            'project_template': self.projectTemplate,

        }

    def getPatterns(
            self,
            file_id: Union[int, List[int], None] = None,
            pattern_id: Union[int, List[int], None] = None,
            pattern_set_id: Union[int, List[int], None] = None,
            series: Union[AnyStr, List[AnyStr], None] = None,
            user_id: Union[int, List[int], None] = None) -> pd.DataFrame:
        """Returns a dataframe of patterns for this project, optionally filtered."""

        # Prepare input
        if not isinstance(file_id, List) and file_id is not None:
            file_id = [file_id]
        if not isinstance(pattern_id, List) and pattern_id is not None:
            pattern_id = [pattern_id]
        if not isinstance(pattern_set_id, List) and pattern_set_id is not None:
            pattern_set_id = [pattern_set_id]
        if not isinstance(series, List) and series is not None:
            series = [series]
        if not isinstance(user_id, List) and user_id is not None:
            user_id = [user_id]

        # Query
        q = models.Pattern.query

        # Filter query as necessary
        if file_id is not None:
            q = q.filter(models.Pattern.file_id.in_(file_id))
        if pattern_id is not None:
            q = q.filter(models.Pattern.pattern_id.in_(pattern_id))
        if pattern_set_id is not None:
            q = q.filter(models.Pattern.pattern_set_id.in_(pattern_set_id))
        if series is not None:
            q = q.filter(models.Pattern.series.in_(series))
        if user_id is not None:
            q = q.filter(models.Pattern.user_id.in_(user_id))

        # Return the dataframe
        return patternDataFrame(q.filter(models.Pattern.project_id == self.id).all())


    def getPatternSet(self, id) -> Optional[PatternSet]:
        """
        Get project's pattern set by ID.
        :return: the PatternSet instance belonging to the id, or None if not found
        """
        return self.patternsets[id] if id in self.patternsets else None

    def getPatternSets(self) -> Dict[int, PatternSet]:
        """
        Get project's pattern sets.
        :return: a dict of the project's PatternSet instances, indexed by id
        """
        return self.patternsets.copy()

    def getTotalPatternCount(self) -> int:
        """
        Get total count of patterns in all the project's pattern sets
        :return: number of patterns
        """
        return sum([ps.count for ps in self.patternsets.values()])

    def listFiles(self) -> List[List[str]]:
        """
        Returns list of files for the project (ID, filename, file path, downsample path).
        :return: list of lists
        """
        return [[f.id, f.name, str(f.origFilePathObj), str(f.procFilePathObj)] for f in self.files]

    def listPatternSets(self) -> List[List[str]]:
        """
        Returns list of pattern sets (ID, names).
        :return: list of l:return:
        """
        return [[ps.id, ps.name] for ps in self.patternsets.values()]

    def loadPatternSets(self) -> None:
        """Load or reload the project's pattern sets."""

        # Reset pattern sets container
        self.patternsets = {}

        # Load pattern sets
        for psm in models.PatternSet.query.filter_by(project_id=self.id).all():
            self.patternsets[psm.id] = PatternSet(self, psm)

    def loadProjectFiles(self, processNewFiles=True):
        """Load or reload files belonging to the project, and process new files if desired."""

        # Reset files list
        self.files = []

        # Will hold all project files that exist in the database (in order to
        # detect new files to process).
        existingFilePathObjs = []

        # Get all of the project's files listed in the database
        fileDBModels = models.File.query.filter_by(project_id=self.id).all()

        # For each project file in the database...
        for fileDBModel in fileDBModels:

            # Verify the original file exists on the file system
            origFilePathObj = Path(fileDBModel.path)
            if origFilePathObj.exists():
                existingFilePathObjs.append(origFilePathObj)
            else:
                logging.error(f"File ID {fileDBModel.id} in the database is missing the original file on the file system at {fileDBModel.path}")
                continue
                # TODO(gus): Do something else with this? Like set error state in the database entry and display in GUI?

            # Verify the processed file exists on the file system
            procFilePathObj = self.processedDirPathObj / getProcFNFromOrigFN(origFilePathObj)
            if not procFilePathObj.exists():
                logging.error(f"File ID {fileDBModel.id} in the database is missing the processed file on the file system at {procFilePathObj}")
                #continue
                # TODO(gus): Do something else with this? Like set error state in the database entry and display in GUI?

            # Instantiate the file class, and attach to this project instance
            self.files.append(File(self, fileDBModel.id, origFilePathObj, procFilePathObj))

        # If processNewFiles is true, then go through and process new files
        if processNewFiles:

            # Get all existing absolute file paths as strings
            existingFilePathStrings = {str(p.resolve()): True for p in existingFilePathObjs}
                
            # For each new project file which does not exist in the database...
            for newOrigFilePathObj in self.originalsDirPathObj.iterdir():

                # Ensure that the path is absolute
                newOrigFilePathObj = newOrigFilePathObj.resolve()

                try:

                    # Check that the file exists (can happen in the case of a dead symlink)
                    if not newOrigFilePathObj.exists():
                        continue

                    # Skip if not file or not .h5
                    if not newOrigFilePathObj.is_file() or newOrigFilePathObj.suffix != '.h5':
                        continue

                    # Skip if matches any already-loaded files
                    if str(newOrigFilePathObj) in existingFilePathStrings:
                        continue

                    # Establish the path of the new processed file
                    newProcFilePathObj = self.processedDirPathObj / getProcFNFromOrigFN(newOrigFilePathObj)

                    # Now that the processing has completed (if not, an exception
                    # would have been raised), add the file to the database and
                    # update the file class instance ID.
                    newFileDBEntry = models.File(project_id=self.id, path=str(newOrigFilePathObj))
                    models.db.session.add(newFileDBEntry)
                    models.db.session.commit()

                    # Add a File class instance for this new file to the files list for this project
                    self.files.append(File(self, newFileDBEntry.id, newOrigFilePathObj, newProcFilePathObj))

                # Handle Ctrl-C
                except KeyboardInterrupt:
                    raise

                # Handle all other exceptions by logging and continuing
                except:
                    logging.error(f"Error loading new file: {traceback.format_exc()}")
        
        # Sort files by filename
        self.files.sort(key=lambda f: f.origFilePathObj.name)

    def setName(self, name):
        """Rename the project."""
        self.model.name = name
        models.db.session.commit()
        self.name = name



from pathlib import Path
from sqlalchemy import distinct, or_, select
import logging
import time
import traceback
import pandas as pd

import audata

from . import models
from .cylib import generateThresholdAlerts
from .series import Series, simpleSeriesName
from .shared import annotationOrPatternOutput

class File:
    """
    Represents a project file. File may operate in file- or realtime-mode. In file-mode, all data is written to & read
    from a file. In realtime-mode, no file is dealt with and instead everything is kept in memory. If no filename
    parameter is passed to the constructor, File will operate in realtime-mode.
    """

    def __init__(self, projparent, id, origFilePathObj, procFilePathObj):

        logging.info(f"\n------------------------------------------------\nACCESSING FILE: (ID {id}) {origFilePathObj}\n------------------------------------------------\n")

        # Holds reference to the project parent
        self.projparent = projparent

        # File ID & paths
        self.id = id
        self.origFilePathObj = origFilePathObj
        self.procFilePathObj = procFilePathObj

        # Store file if already read
        self._file = None
        self._processed_file = None

        # Filename
        self.name = Path(self.origFilePathObj).name

        # Will hold the data series from the file
        self.series = []

    @property
    def f(self):       
        if self._file is None:
            # Open the original file only if the donwsampled version exists
            # TODO(vedant) : does this behavior need to be changed? since current multiprocessing does not support prioritization,
            # users may have to wait for a while to view their files
            self._file = audata.File.open(str(self.origFilePathObj), return_datetimes=False)

            # Load series data into memory
            self.load()

        return self._file

    @property
    def pf(self):
        if self._processed_file is None:
            if not self.procFilePathObj.exists():
                # TODO(vedant/gus) : inform the front end instead of backend about the file being downsampled
                raise IOError("Please wait, file being downsampled")

            else:
                logging.info(f"Opening processed file {self.procFilePathObj}.")
                tmp_file = Path(str(self.procFilePathObj)+'.tmp')
                if tmp_file.exists():
                    raise RuntimeError("Temp file for corresponding processed file does not exist. This indicates that the file is currently in the process of being downsampled, or the downsampled file is corrupted")

                # Loads the processed file if no issues are detected
                self._processed_file = audata.File.open(str(self.procFilePathObj), return_datetimes=False)

        return self._processed_file

    def close(self):
        """ Use this for closing original and processed .h5 without removing File object from memory"""

        logging.info("Closing original and processed files")

        # Close the original file
        try:
            self._file.close()
        except:
            pass

        # Close the processed file
        try:
            self._processed_file.close()
        except:
            pass

        self._file, self._processed_file = None, None

    def __del__(self):
        self.close()

    def addSeriesData(self, seriesData):
        """
        Takes new data to add to one or more data series for the file (currently
        works only in realtime-mode). The new data is assumed to occur after any
        existing data. The parameter, seriesData, should be a dict of dicts of
        lists as follows:

          {
            'seriesName': {
              'times': [ t1, t2, ... , tn ],
              'values': [ v1, v2, ... , vn ]
            }, ...
          }
        :param seriesData: dict of dicts
        :return: updated file data for transmission to realtime subscribers
        """

        logging.info(f"Adding series data: {seriesData}")

        if self.mode() != 'realtime':
            raise Exception('Can only addSeriesData() while in mem-mode.')

        if not isinstance(seriesData, dict):
            raise Exception('Invalid seriesData parameter received for addSeriesData().')

        # This will hold the update data to return
        update = {
            'series': {}
        }

        for s in seriesData:

            if not isinstance(seriesData[s], dict):
                raise Exception('Invalid seriesData[s] parameter received for addSeriesData().')

            # Retrieve or create the series
            series = self.getSeriesOrCreate(s)

            # If that failed, raise an exception
            if series is None:
                raise Exception('Attempting to create or retrieve series failed:', s)

            # Add the new series data
            series.addData(seriesData[s])

            nones = [None] * len(seriesData[s]['times'])

            update['series'][s] = {
                "id": s,
                "labels": ['Date/Offset', 'Min', 'Max', simpleSeriesName(s)],
                "data": [list(i) for i in zip(seriesData[s]['times'], nones, nones, seriesData[s]['values'])],
                "output_type": 'real'
            }

        logging.info(f"Adding series data completed. Returning {update}")

        return update

    def createAnnotation(self, user_id, left=None, right=None, top=None, bottom=None, seriesID='', label='', pattern_id=None):
        """Create an annotation for the file"""

        # If the pattern_id is set, determine the pattern_set_id.
        pattern_set_id = None;
        if pattern_id:
            pattern_set_id = models.Pattern.query.filter_by(id=pattern_id).first().pattern_set_id

        newann = models.Annotation(
            user_id=user_id,
            project_id=self.projparent.id,
            file_id=self.id,
            series=seriesID,
            left=left,
            right=right,
            top=top,
            bottom=bottom,
            label=label,
            pattern_set_id=pattern_set_id,
            pattern_id=pattern_id
        )
        models.db.session.add(newann)
        models.db.session.commit()

        return newann.id

    def deleteAnnotation(self, user_id, id):
        """Deletes the annotation with the given ID, after some checks. Returns true or false to indicate success."""

        # Get the annotation in question
        annotationToDelete = models.Annotation.query.filter_by(id=id).first()

        # Verify the user has requested to delete his or her own annotation.
        if annotationToDelete.user_id != user_id:
            logging.critical(
                f"Error/securityissue on annotation deletion: User (id {user_id}) tried to delete an annotation (id {id}) belonging to another user (id {annotationToDelete.user_id}).")
            return False

        # Having reached this point, delete the annotation
        models.db.session.delete(annotationToDelete)
        models.db.session.commit()

        # Return true to indicate success
        return True

    # TODO(gus): Turn this into DataFrame output in line with Project.detectPatterns.
    def detectPatterns(
            self,
            type,
            series,
            thresholdlow=None,
            thresholdhigh=None,
            duration=None,
            persistence=None,
            maxgap=None,
            expected_frequency=0,
            min_density=0,
            drop_values_below=None,
            drop_values_above=None,
            drop_values_between=None,
        ):

        # Ensure that the file is open
        _ = self.f

        assert isinstance(series, str) or isinstance(series, list), "series must be a string or list"
        
        assert type == 'patterndetection', "detectPatterns() currently only supports type='patterndetection'"

        assert thresholdlow is not None or thresholdhigh is not None, "We need at least one of thresholdlow or thresholdhigh in order to perform pattern detection."

        assert duration is not None, "We need a duration in order to perform pattern detection."

        assert persistence is not None, "We need a persistence in order to perform pattern detection."

        assert maxgap is not None, "We need a maxgap in order to perform pattern detection."

        # If series is a string, make it a list
        if isinstance(series, str):
            series = [series]

        # Determine the mode (see generateThresholdAlerts function description for details on this parameter).
        if thresholdhigh is None:
            mode = 0
            thresholdhigh = 0
        elif thresholdlow is None:
            mode = 1
            thresholdlow = 0
        else:
            mode = 2
            
        if type == 'patterndetection':

            # Find the series & run pattern detection
            patterns = []
            for ts in series:
                for s in self.series:
                    if s.id == ts:
                        patterns += s.generateThresholdAlerts(
                            thresholdlow, 
                            thresholdhigh, 
                            mode, duration, 
                            persistence, 
                            maxgap, 
                            expected_frequency=expected_frequency, 
                            min_density=min_density,
                            drop_values_below=drop_values_below,
                            drop_values_above=drop_values_above,
                            drop_values_between=drop_values_between,
                        ).tolist()
                        continue

            return patterns

        # elif type == 'correlation':

        #     # TODO(Gus): TEMP!
        #     series2 = '/data/numerics/HR.HR:value'
        #     timewindow='10min'

        #     # The series2 parameter is required for correlation detection.
        #     if series2 is None:
        #         print("Error: Correlation detection requires series2 parameter.")
        #         return []

        #     # Find the series
        #     for s in self.series:
        #         if s.id == series:
        #             s1 = s
        #         if s.id == series2:
        #             s2 = s

        #     t1, v1 = s1.pullRawDataIntoMemory(returnValuesOnly=True)
        #     t2, v2 = s2.pullRawDataIntoMemory(returnValuesOnly=True)

        #     s1Data = pd.DataFrame({
        #         'time': t1,
        #         'val1': v1
        #     })
        #     s2Data = pd.DataFrame({
        #         'time': t2,
        #         'val2': v2
        #     })

        #     # Assemble into series 1 & series 2 dataframes
        #     # s1Data = pd.concat((t1, v1), axis=1)
        #     # s2Data = pd.concat((t2, v2), axis=1)
        #     #
        #     # # Assert column names
        #     # s1Data.columns = ['time', 'val1']
        #     # s2Data.columns = ['time', 'val2']

        #     # Create an inner joined dataset with common time column
        #     df = s1Data.merge(s2Data, on='time')

        #     # Filter out null values
        #     df = df[df.notnull()]

        #     # Create a datetime column
        #     df['time_conv'] = pd.to_datetime(df.time, unit='s')

        #     # Set the index to the datetime column
        #     df.set_index('time_conv', inplace=True)

        #     logging.debug(df)

        #     # Generate the rolling-window correlation
        #     corr = df['val1'].rolling(timewindow).corr(other=df['val2'])

        #     # Now run pattern detection on the rolling-window correlation
        #     alerts = generateThresholdAlerts(df['time'].to_numpy(), corr.to_numpy(), thresholdlow, thresholdhigh, mode, duration, persistence, maxgap).tolist()

        #     return alerts

        # Having reached this point, we were unable to generate the alerts.
        return []

    def getEvents(self):
        """Returns all event series"""

        logging.info(f"Assembling all event series for file {self.origFilePathObj}.")
        start = time.time()

        events = {}

        # If we're in realtime-mode, we have no events to return, so return now.
        if self.mode() == 'realtime':
            return events

        def catcols(row, idcol=None):
            idx = row[idcol]
            return (idx, '\n'.join(
                [f'{row[c]}' for c in row.index \
                    if (c is not idcol) and (row[c] is not None) and (row[c] != '')]))

        try:

            # Prepare references to HDF5 data
            meds = self.f['ehr/medications'][:]
            events['meds'] = meds.apply(catcols, 1, idcol='time').tolist()

        except Exception:
            print("Error retrieving meds.")
            traceback.print_exc()

        try:

            # Prepare references to HDF5 data
            ce = self.f['low_rate'][:]
            events['ce'] = ce.apply(catcols, 1, idcol='date').tolist()

        except Exception:
            print("Error retrieving ce.")
            traceback.print_exc()

        end = time.time()
        logging.info(f"Completed assembly of all event series for file {self.origFilePathObj}. Took {str(round(end - start, 5))}s.")

        return events

    def _loadonstart(self):
        while True:
            # Open the original file
            self.file = audata.File.open(str(self.origFilePathObj), return_datetimes=False)

            # Load the processed file if it exists
            if self.procFilePathObj.exists():
                logging.info(f"Opening processed file {self.procFilePathObj}.")
                self.processed_file = audata.File.open(str(self.procFilePathObj), return_datetimes=False)
                self.pf_open = True

                # If we've been asked only to generate the processed file and it already exists, return now.
                if self.processOnly:
                    break

            # Load series data into memory
            self.load()
            self.f_open = True

            # If the processed file does not exist and we're supposed to process
            # new file data, process it.
            if not self.procFilePathObj.exists() and self.processNewFiles:
                logging.info(f"Generating processed file for {self.origFilePathObj}")
                self.process()
                self.pf_open = True

            # If the processed file still does not exist, raise an exception
            if not self.procFilePathObj.exists():
                raise Exception(f"File {self.origFilePathObj} has not been processed and therefore cannot be loaded.")
            
            break
    

    def getInitialPayload(self, user_id):
        """Produces JSON output for all series in the file at the maximum time range."""

        logging.info(f"Assembling all series full output for file {self.origFilePathObj}.")
        start = time.time()

        if self.mode() == 'file':
            try:
                _ = self.pf
            except:
                return {}

        # Assemble the output object.
        outputObject = {
            'filename': self.origFilePathObj.name,
            'annotationsets': [{
                'id': 'general',
                'name': 'General',
                'description': None,
                'annotations': [annotationOrPatternOutput(a) for a in models.Annotation.query.filter_by(user_id=user_id, project_id=self.projparent.id, file_id=self.id, pattern_set_id=None).all()],
                'show': True,
            }] + [
                {
                    'id': patternset.id,
                    'name': patternset.name,
                    'description': patternset.description,
                    'annotations': [annotationOrPatternOutput(a) for a in models.Annotation.query.filter_by(user_id=user_id, project_id=self.projparent.id, file_id=self.id, pattern_set_id=patternset.id).all()],
                    'show': patternset.show_by_default,
                } for patternset in models.PatternSet.query.filter(models.PatternSet.project_id==self.projparent.id, or_(
                    models.PatternSet.id.notin_(
                        select(distinct(models.patternSetAssignments.c.pattern_set_id))
                    ),
                    models.PatternSet.id.in_(
                        select(models.patternSetAssignments.c.pattern_set_id).where(models.patternSetAssignments.c.user_id==user_id)
                    )
                )).all()
            ],
            'patternsets': [
                {
                    'id': patternset.id,
                    'name': patternset.name,
                    'description': patternset.description,
                    'patterns': [annotationOrPatternOutput(pattern) for pattern in models.Pattern.query.filter_by(pattern_set_id=patternset.id, file_id=self.id).all()],
                    'show': patternset.show_by_default,
                } for patternset in models.PatternSet.query.filter(models.PatternSet.project_id==self.projparent.id, or_(
                    models.PatternSet.id.notin_(
                        select(distinct(models.patternSetAssignments.c.pattern_set_id))
                    ),
                    models.PatternSet.id.in_(
                        select(models.patternSetAssignments.c.pattern_set_id).where(models.patternSetAssignments.c.user_id==user_id)
                    )
                )).all()
            ],
            'baseTime': 0, # ATW: Not sure if this is still necessary.
            'events': self.getEvents(),
            'metadata': self.getMetadata(),
            'series': {}
        }

        for s in self.series:
            outputObject['series'][s.id] = s.getFullOutput()

        end = time.time()
        logging.info(f"Completed assembly of all series full output for file {self.origFilePathObj}. Took {str(round(end - start, 5))}s.")

        # Return the output object
        return outputObject

    def getMetadata(self):
        """Returns a dict of file metadata."""
        try:
            return self.f.file_meta
        except:
            return {}

    def getSeries(self, seriesid):
        """
        Returns the series instance corresponding to the provided series ID, or None if the series cannot be found.
        The seriesid format is [full_series_path]:[value_column], e.g. /data/numerics/HR:value.
        """
        logging.info(f"Doing a File.getSeries() on {seriesid}")

        for s in self.series:
            if s.id == seriesid:
                return s
        return None

    def getSeriesNames(self):
        """Returns a list of series names available in the file."""

        seriesNames = []

    def getSeriesOrCreate(self, seriesid):
        """Retreves or creates & returns the series corresponding to the provided series ID."""

        # Attempt to retrieve the series
        series = self.getSeries(seriesid)

        # Otherwise, create the series
        if series is None:

            # Create new series
            self.series.append(Series([seriesid], 'time', 'value', self))

            # Retrieve the newly-created series
            series = self.getSeries(seriesid)

        return series

    def getSeriesRangedOutput(self, seriesids, start, stop):
        """Produces JSON output for a given list of series in the file at a specified time range."""

        logging.info(f"Assembling series ranged output for file {self.origFilePathObj}, series [{', '.join(seriesids)}].")
        st = time.time()

        outputObject = {
            'series': {}
        }

        for s in self.series:
            if s.id in seriesids:
                outputObject['series'][s.id] = s.getRangedOutput(start, stop)

        et = time.time()
        logging.info(f"Completed assembly of series ranged output for file {self.origFilePathObj}, series [{', '.join(seriesids)}]. Took {str(round(et - st, 5))}s.")

        # Return the output object
        return outputObject

    def load(self):
        """
        Loads the necessary data into memory for an already-processed data file
        (does not load data though). Sets up classes for all series from file but
        does not load series data into memory).
        """

        logging.info('Loading series from file.')

        # Iterate through all datasets in the partial file
        for (ds, _) in self.f.recurse():
            self.loadSeriesFromDataset(ds)

        logging.info('Completed loading series from file.')

    def loadSeriesFromDataset(self, ds):
        """Load all available series from a dataset"""

        # Grab the columns.
        cols = ds.columns

        # Determine the time column name. Look for a column names 'time', 'timestampe',
        # or the first time column, in that order.
        timecol = None
        if 'time' in cols:
            timecol = 'time'
        elif 'timestamp' in cols:
            timecol = 'timestamp'
        else:
            for col in cols:
                if cols[col]['type'] == 'time':
                    timecol = col
                    break

        # If time column not available, print an error & return
        if timecol is None:
            logging.error(f"Did not find a time column in {self.origFilePathObj}.\nCols: {cols}")
            return
        else:
            logging.info(f'Using time column {ds.name}:{timecol}')

        # Iterate through all remaining columns and instantiate series for each
        for valcol in (c for c in cols if c != timecol):
            coltype = cols[valcol]['type']
            if coltype in ('real', 'integer'):
                logging.info(f'  - Adding {coltype} series: {ds.name}:{valcol}')
                self.series.append(Series(ds, timecol, valcol, self))
            else:
                logging.warning(f'  - Skipping unsupported {coltype} series: {valcol}')

    # TODO(gus): When reviving realtime functionality, revise this
    def mode(self):
        """Returns the mode in which File is operating, either "file" or "realtime"."""
        return 'file'

    def process(self):
        """Process and store all downsamples for all series for the file."""

        # Create a path name for temporary file
        tmp_file = self.procFilePathObj.with_suffix(self.procFilePathObj.suffix + '.tmp')
        try:

            logging.info(f"Processing & storing all series for file {self.origFilePathObj}.")
            start = time.time()

            # Print user message
            print(f"Downsampling file {self.name}...")

            # Create the file for storing processed data.
            self._processed_file = audata.File.new(str(self.procFilePathObj), overwrite=False, time_reference=self.f.time_reference, return_datetimes=False)
            
            # Create a tmp file to indicate a file that has in the process of getting donwsampled
            # These tmp files will be deleted either after successful downsampling or after restarting
            # the viewer. 
            #
            # Tmp files are always presereved in the event where the file could not be successfully downsampled
            # and the processed file could not be deleted.
            with open(str(tmp_file), 'w') as fp:
                pass

            # Process & store numeric series
            for s in self.series:
                s.processAndStore()

            self._processed_file.flush()

            # Reload series data in memory since we have new downsamples
            self.load()

            # Print user message
            print("Done.")

            end = time.time()
            logging.info(f"Completed processing & storing all series for file {self.origFilePathObj}. Took {str(round((end - start) / 60, 3))} minutes).")

        except (KeyboardInterrupt, SystemExit):

            logging.warning("Interrupt detected. Aborting as requested.")
            logging.warning(f"Deleting partially completed processed file {self.procFilePathObj}.")

            # Close the file
            try:
                self._processed_file.close()
            except Exception as e:
                logging.error(f"Unable to close the processed file.\n{e}\n{traceback.format_exc()}")

            # Delete the processed data file
            try:
                self.procFilePathObj.unlink()
            except Exception as e:
                logging.error(f"Unable to delete file successfully. \n{e}\n{traceback.format_exc()}")
            else:
                tmp_file.unlink()
    

            # Quit the program
            quit()

        except Exception as e:

            logging.error(f"There was an exception while processing & storing data for {self.origFilePathObj}.\n{e}\n{traceback.format_exc()}\nDeleting partially completed processed file {self.procFilePathObj}.")

            # Close the file
            try:
                self._processed_file.close()
            except Exception as e:
                logging.error(f"Unable to close the processed file.\n{e}\n{traceback.format_exc()}")

            # Delete the processed data file
            try:
                self.procFilePathObj.unlink(missing_ok=True)
            except Exception as e:
                logging.error(f"Unable to delete file successfully. \n{e}\n{traceback.format_exc()}")
            else:
                tmp_file.unlink()

            logging.info("File has been removed.")

            # Re-raise the exception
            raise

        else:
            # Deletes temporary files if files are downsampled successfully
            tmp_file.unlink()

    def updateAnnotation(self, user_id, id, left=None, right=None, top=None, bottom=None, seriesID='', label=''):
        """Update an annotation with new values"""

        # Get the annotation in question
        annotationToUpdate = models.Annotation.query.filter_by(id=id).first()

        # Verify the user has requested to update his or her own annotation.
        if annotationToUpdate.user_id != user_id:
            logging.critical(f"Error/security issue on annotation update: User (id {user_id}) tried to update an annotation (id {id}) belonging to another user (id {annotationToUpdate.user_id}).")
            return False

        # Set updated values
        annotationToUpdate.series = seriesID
        annotationToUpdate.left = left
        annotationToUpdate.right = right
        annotationToUpdate.top = top
        annotationToUpdate.bottom = bottom
        annotationToUpdate.label = label

        # Commit the changes
        models.db.session.commit()

        # Return true to indicate success
        return True



import logging
import psutil
import time

from .config import config
from .cylib import buildDownsampleFromRaw, buildNextDownsampleUp, getSliceParam, numDownsamplesToBuild

# Represents a set of downsamples for a series of data.
class DownsampleSet:

    def __init__(self, seriesparent):

        # Set the series parent
        self.seriesparent = seriesparent

        # Holds the number of downsamples available for the series
        self._numDownsamples = None

    @property
    def numDownsamples(self):
        if self._numDownsamples is None:
            try:
                # Attempt to access the processed file
                _ = self.seriesparent.fileparent.pf
            except:
                # If we received an exception (which we assume indicates the processed file does not exist),
                # return 0 but do not set self._numDownsamples
                return 0

            # Otherwise, we're able to access the processed file, so go ahead and set self._numDownsamples
            self._numDownsamples = self.getNumDownsamplesFromFile()

        return self._numDownsamples

    # Returns the full series output at the highest downsample level, or None
    # if no downsample exists for the series.
    def getFullOutput(self):

        if self.numDownsamples < 1:
            return None

        return self.seriesparent.fileparent.pf['/'.join(self.seriesparent.h5pathDownsample) + '/' + '0'][:]

    # Returns the number of downsamples available for this series in the
    # processed data file.
    def getNumDownsamplesFromFile(self):

        grp = self.seriesparent.fileparent.pf['/'.join(self.seriesparent.h5pathDownsample)]

        # If no downsamples are available, return 0
        if grp is None:
            return 0

        # Get the keys present in the group (they should all be strings of
        # integers 0..n where n+1 is the count of downsamples.
        keys = grp.list()['datasets']

        # We start with -1
        maxDownsampleIndex = -1

        # Crawl up the downsample indicates, confirming that each is present.
        # There's obviously an easier way to do this, but we're doing some
        # verification of integrity here.
        while str(maxDownsampleIndex + 1) in keys:
            maxDownsampleIndex = maxDownsampleIndex + 1

        # We have the real max downsample index, so that plus one is the count.
        return maxDownsampleIndex + 1

    # Returns the number of intervals for the downsample at index i.
    def getNumIntervalsByIndex(self, i, nds=-1):

        if nds == -1:
            nds = self.numDownsamples

        # Handle negative index
        if i < 0:
            i = nds + i

        return config['M'] * (config['stepMultiplier'] ** i)

    # Returns a slice of the appropriate downsample for the given time range, or
    # nothing if there is no appropriate downsample available (in this case, raw
    # data should be used). Expects starttime & stoptime to be time offsets
    # floats in seconds.
    def getRangedOutput(self, starttime, stoptime):

        # If there are no downsamples available, we cannot provide one
        if self.numDownsamples < 1:
            return None

        # Calculate the timespan of the view window, in seconds
        timespan = stoptime - starttime

        # Get index of the appropriate downsample to use
        dsi = self.whichDownsampleIndexForTimespan(timespan)

        # If we should be using raw data, return nothing
        if dsi == -1:
            return None

        # Get reference to the downsample dataset in the processed file
        ds = self.seriesparent.fileparent.pf['/'.join(self.seriesparent.h5pathDownsample) + '/' + str(dsi)]

        # Find the start & stop indices based on the start & stop times.
        startIndex = getSliceParam(ds, '0', 0, starttime)
        stopIndex = getSliceParam(ds, '0', 1, stoptime)

        # Return the downsample slice
        return ds[startIndex:stopIndex]

    # Returns the time-per-interval for the downsample at index i.
    def getTimePerIntervalByIndex(self, i, nds=-1):

        if nds == -1:
            nds = self.numDownsamples

        # Handle negative index
        if i < 0:
            i = nds + i

        return self.seriesparent.rd.timespan / self.getNumIntervalsByIndex(i, nds)

    # Build all necessary downsamples, and store in the processed file.
    def processAndStore(self):

        p = psutil.Process()

        # We assume:
        #   - Datagroup has two non-empty datasets, "datetime" and "value";
        #   - The two datasets are of equal length;
        #   - There exists a file for storing processed data.
        if \
                (len(self.seriesparent.rawTimes) < 1 or len(self.seriesparent.rawValues) < 1) or \
                len(self.seriesparent.rawTimes) != len(self.seriesparent.rawValues) or \
                not hasattr(self.seriesparent.fileparent, 'pf'):
            return

        # Get an array of the downsamples to build (each element of the array
        # is a number of intervals to divide the data set into).
        ndtb = numDownsamplesToBuild(self.seriesparent.rawTimes, config['M'], config['stepMultiplier'])

        # If we do not need to build any downsamples, return now.
        if ndtb < 1:
            return

        # Begin by building the last downsample from the raw data first. Then
        # proceed by building the next downsample up from the previously-built
        # downsample until all downsamples have been created.
        logging.info(f"MEM PRE-DSBLD: {p.memory_full_info().uss / 1024 / 1024} MB")
        for i in range(-1, -ndtb - 1, -1):

            logging.info(f"Creating downsample ({self.getNumIntervalsByIndex(i, ndtb)} intervals possible).")
            start = time.time()

            # Build the downsample
            if i == -1:
                logging.info("This pass from raw.")
                downsample = buildDownsampleFromRaw(self.seriesparent.rawTimes, self.seriesparent.rawValues, self.getNumIntervalsByIndex(i, ndtb))
            else:
                downsample = buildNextDownsampleUp(previousDownsample, self.getTimePerIntervalByIndex(i + 1, ndtb), config['stepMultiplier'])

            logging.info(f"MEM AFT-DSBLD: {p.memory_full_info().uss / 1024 / 1024} MB")

            # Save the just-computed downsample for use on the next loop iteration
            previousDownsample = downsample

            logging.info(f"MEM AFT-REPLP: {p.memory_full_info().uss / 1024 / 1024} MB")

            end = time.time()
            logging.info(f"Done creating downsample. Yielded {downsample.shape[0]} actual intervals. Took {round(end - start, 5)}s.")

            logging.info("Storing the downsample to the processed file.")
            start = time.time()

            try:

                # Store the downsample
                dds_name = '{}/{}'.format('/'.join(self.seriesparent.h5pathDownsample), i % ndtb)
                self.seriesparent.fileparent.pf[dds_name] = downsample
                logging.info(f"MEM AFT-STRFL: {p.memory_full_info().uss / 1024 / 1024} MB")

            except:

                logging.info(f"There was an exception while creating the dataset in the processed data file at the path: {'/'.join(self.seriesparent.h5pathDownsample)}/{i%ndtb}.")
                raise

            end = time.time()
            logging.info(f"Done storing to file. Took {round(end - start, 5)}s.")

        # Clear self._numDownsamples so that it updates the next time it's accessed
        self._numDownsamples = None

    # Returns the index of the appropriate downsample which should be used for
    # the given timespan, or -1 if raw data should be used instead.
    def whichDownsampleIndexForTimespan(self, timespan):

        # Calculate the largest interval size to deliver <= M intervals, in seconds
        maxTimePerInterval = timespan / config['M']

        # Determine which downsample to use. To do this, find the first
        # downsample, going from the largest interval downwards, which is too
        # small and then select the previous one.
        i = 0
        while i < self.numDownsamples and self.getTimePerIntervalByIndex(i) >= maxTimePerInterval:
            i = i + 1
        dsi = i - 1

        # We do not expect this, but in case even the largest interval size was
        # too small, return the largest one we have.
        if dsi < 0:
            dsi = 0

        # If we reached the lowest downsample level, check whether we should be
        # using real data points. If so, return -1.
        if (dsi == self.numDownsamples - 1) and maxTimePerInterval <= self.getTimePerIntervalByIndex(dsi) / 2:
            return -1

        return dsi



from abc import ABC, abstractmethod
from typing import AnyStr, List, Dict, Optional, Union
from functools import partial

class AdvancedFeaturizer(ABC):
    pass

class SimpleFeaturizer(ABC):

    id = None
    parameters = []

    # def __init__(self, files: Union[str, List[str]], series: Union[str, List[str], None]):
    #
    #     # If a string was provided as the files param (i.e. referencing a single file), convert it to a list
    #     if isinstance(files, str):
    #         files = [files]
    #
    #     # If a string was provided as the series param (i.e. referencing a single series), convert it to a list
    #     if isinstance(series, str):
    #         series = [series]
    #
    #     # Set files & series as instance properties
    #     self.files = files
    #     self.series = series

    def __init__(self):
        if self.id is None:
            raise Exception("Featurizer has no ID.")
        if self.name is None:
            raise Exception(f"Featurizer {self.id} has no name.")

    def getFeaturizeFunction(self, params):
        # TODO(gus): Wrap iteratively in try/except?
        print(f"Params pre-process: {params}")
        preparedParams = self.prepareParams(params)
        print(f"Params post-process: {preparedParams}")
        return partial(self.featurize, params=preparedParams)

    def getFields(self):
        """
        Produces and returns a list of field dicts ready to be marshalled to JSON and supplied to Webix as a JavaScript
        array representing the featurizer's form fields.
        """
        return [field.getField() for field in self.parameters]

    @abstractmethod
    def featurize(self, data, params={}, series=[]):
        """
        This will be the function which, given a Pandas DataFrame and parameters dict, featurizes and returns the
        feature's scalar value output.
        :param data: the dataframe of data
        :param params: a dict of key=>value pairs of the parameters
        :param series: list of references to entire series/dataframes (irrespective of rolling window operation), with the target series in the first position
        :return: Scalar value output of featurization
        """
        raise NotImplementedError("Error! Required featurizer method 'featurize' not implemented.")

    def prepareParams(self, params):
        """
        Returns a processed version of params given web form input.
        """
        newParams = {}
        for p in self.parameters:

            try:

                val = params[p.id]
                if p.data_type == 'boolean':
                    if val is None or val=='':
                        val = p.default
                    elif not val or val==0 or val=='0' or val=='false' or val=='False':
                        val = False
                    else:
                        val = True
                else:
                    if val is None or val=='':
                        val = p.default

                newParams[p.id] = val

            except:
                newParams[p.id] = p.default

        return newParams


class FeaturizerParameter():

    def __init__(self,
                 id,
                 name,
                 description='',
                 data_type='float',
                 form_field_type='input',
                 options=[],
                 required=True,
                 default=None,
                 ):
        """
        :param id: Parameter ID
        :param name: Name of the parameter
        :param description: Description of the parameter with usage instructions for users
        :param data_type: Indicates the data type of the parameter. Possible values are 'string', 'int', and 'float'.
        :param form_field_type: Indicates the type of form field to present to the user. Possible values are 'input',
        'dropdown', 'textarea', 'checkbox', 'radio', and 'slider'. Default value is 'input'. If data_type is 'boolean',
        then form_field_type and options are ignored.
        :param options: Dict (option value=>display value) of options available for user to choose from, if the
        form_field_type is 'dropdown'. If data_type is 'boolean', then form_field_type and options are ignored.
        :param required: Boolean indicating whether the parameter must be specified by the user.
        :param default: Specifies the default value to use for the parameter if a value is not specified by the user.
        :return: None
        """

        # Validation dictionaries
        valid_data_types = ['string', 'int', 'float', 'boolean']
        valid_form_field_types = ['input', 'dropdown', 'textarea', 'checkbox', 'radio', 'slider']

        # Validate properties
        if data_type not in valid_data_types:
            raise Exception(f"Invalid data_type provided: {data_type}")
        if form_field_type not in valid_form_field_types:
            raise Exception(f"Invalid form_field_type provided: {form_field_type}")
        if not len(name) > 0:
            raise Exception(f"Empty parameter name provided: {name}")

        # Boolean-specific properties enforcement
        if data_type == 'boolean':

            if required:
                form_field_type = 'checkbox'
                options = []
            else:
                form_field_type = 'dropdown'
                options = {'': '', 'true': 'True', 'false': 'False'}

            if not default:
                default = False
            else:
                default = True
            # TODO: TEMP!!!!!
            default = 'default'

        # Set properties
        self.id = id
        self.name = name
        self.description = description
        self.data_type = data_type
        self.form_field_type = form_field_type
        self.options = options
        self.required = required
        self.default = default

    def getField(self):
        """
        This function produces and returns a dict ready to be marshalled to JSON and
        then supplied to Webix as a JavaScript object representing a form field.
        """

        field = {
            'name': self.id,
            'id': self.id,
            #'labelWidth': 230,
            'labelWidth': 'auto',
            'labelAlign': 'left',
            'inputAlign': 'right',
            'label': self.name + ('*' if self.required else ''),
            'tooltip': self.description + (f"\n\nDefault value: {self.default}" if self.default is not None else ''),
        }

        if self.form_field_type in ('radio', 'textarea', 'checkbox'):
            field['view'] = self.form_field_type
            if self.default is not None:
                field['value'] = self.default
        elif self.form_field_type == 'dropdown':
            field['view'] = 'select'
        elif self.form_field_type == 'input':
            field['view'] = 'text'
        elif self.form_field_type == 'slider':
            raise NotImplementedError("Slider is not implemented.")
        else:
            raise Exception("Form field type not recognized.")

        if self.form_field_type in ('dropdown', 'radio'):
            field['options'] = [{'id': key, 'value': value} for key, value in self.options.items()]

        return field



from flask import current_app
from .tst_app import Role

def test_roles(db):
    um =  current_app.user_manager
    hashed_password = um.hash_password('Password1')
    User = um.db_adapter.UserClass

    # create users and roles
    role1 = Role(name='Role 1')
    role2 = Role(name='Role 2')

    # user0 has no roles
    user0 = User(username='user0', email='user0@example.com', password=hashed_password, active=True)
    db.session.add(user0)
    
    # user1 has only role1 
    user1 = User(username='user1', email='user1@example.com', password=hashed_password, active=True)
    user1.roles.append(role1)
    db.session.add(user1)
    
    # user2 has role1 and role2
    user2 = User(username='user2', email='user2@example.com', password=hashed_password, active=True)
    user2.roles.append(role1)
    user2.roles.append(role2)
    db.session.add(user2)
    db.session.commit()

    # test has_role()
    assert user0.has_role('Role 1')==False
    assert user0.has_role('Role 2')==False
    assert user0.has_role('Role 3')==False
    assert user0.has_role('Role 1', 'Role 2')==False
    assert user0.has_role('Role 2', 'Role 1')==False
    assert user0.has_role('Role 1', 'Role 2', 'Role 3')==False

    assert user1.has_role('Role 1')==True
    assert user1.has_role('Role 2')==False
    assert user1.has_role('Role 3')==False
    assert user1.has_role('Role 1', 'Role 2')==True
    assert user1.has_role('Role 2', 'Role 1')==True
    assert user1.has_role('Role 1', 'Role 2', 'Role 3')==True

    assert user2.has_role('Role 1')==True
    assert user2.has_role('Role 2')==True
    assert user2.has_role('Role 3')==False
    assert user2.has_role('Role 1', 'Role 2')==True
    assert user2.has_role('Role 2', 'Role 1')==True
    assert user2.has_role('Role 1', 'Role 2', 'Role 3')==True

    # test has_roles()
    assert user0.has_roles('Role 1')==False
    assert user0.has_roles('Role 2')==False
    assert user0.has_roles('Role 3')==False
    assert user0.has_roles('Role 1', 'Role 2')==False
    assert user0.has_roles('Role 2', 'Role 1')==False
    assert user0.has_roles('Role 1', 'Role 2', 'Role 3')==False

    assert user1.has_roles('Role 1')==True
    assert user1.has_roles('Role 2')==False
    assert user1.has_roles('Role 3')==False
    assert user1.has_roles('Role 1', 'Role 2')==False
    assert user1.has_roles('Role 2', 'Role 1')==False
    assert user1.has_roles('Role 1', 'Role 2', 'Role 3')==False

    assert user2.has_roles('Role 1')==True
    assert user2.has_roles('Role 2')==True
    assert user2.has_roles('Role 3')==False
    assert user2.has_roles('Role 1', 'Role 2')==True
    assert user2.has_roles('Role 2', 'Role 1')==True
    assert user2.has_roles('Role 1', 'Role 2', 'Role 3')==False

    # delete users and roles




import numpy as np
import datetime as dt

from .cylib import getSliceParam

# Represents raw data for a single time series
class RawData:

    # RawData may operate in file-mode or data-mode. If seriesparent.h5path`
    def __init__(self, seriesparent):

        # Set the series parent
        self.seriesparent = seriesparent

        # Grab a reference to the dataset
        dataset = self.getDatasetReference()

        # Holds the number of data points in the raw data series
        self.len = dataset.nrow

        # Holds the timespan of the time series
        if self.len < 2:
            self.timespan = 0
        else:
            # self.timespan = np.abs(np.diff(dataset[[-1, 0]][self.seriesparent.timecol].values.astype(np.float64))[0])
            self.timespan = dataset[[0, -1]]
            self.timespan = self.timespan[self.seriesparent.timecol]
            self.timespan = self.timespan.values.astype(np.float64)
            self.timespan = np.abs(np.diff(self.timespan)[0])

    # Returns a slice of the appropriate downsample for the given time range, or
    # nothing if there is no appropriate downsample available (in this case, raw
    # data should be used). Expects starttime & stoptime to be time offsets
    # floats in seconds.
    def getRangedOutput(self, starttime, stoptime):

        # Grab a reference to the dataset
        ds = self.getDatasetReference()

        # Find the start & stop indices based on the start & stop times.
        # startIndex = np.searchsorted(self.rawTimeOffsets, starttime)
        # stopIndex = np.searchsorted(self.rawTimeOffsets, stoptime, side='right')
        startIndex = getSliceParam(ds, self.seriesparent.timecol, 0, starttime)
        stopIndex = getSliceParam(ds, self.seriesparent.timecol, 1, stoptime)

        # Slice the output data the output data
        ds_slice = ds[startIndex:stopIndex]

        # Pull the raw times & values
        rawTimes = ds_slice[self.seriesparent.timecol].values.astype(np.float64)
        rawValues = ds_slice[self.seriesparent.valcol].values.astype(np.float64)

        # Drop nan values
        mask = ~np.isnan(rawValues)
        rawTimes = rawTimes[mask]
        rawValues = rawValues[mask]

        nones = [None] * len(rawTimes)

        return [list(i) for i in zip(rawTimes, nones, nones, rawValues)]

    def getDatasetReference(self):
        
        return self.seriesparent.fileparent.f['/'.join(self.seriesparent.h5path)]



"""
    tests.test_invalid_forms.py
    ---------------------------
    Flask-User automated tests:
    Tests all forms with as many invalid field values as possible

    :copyright: (c) 2013 by Ling Thio
    :author: Ling Thio (ling.thio@gmail.com)
    :license: Simplified BSD License, see LICENSE.txt for more details.
"""

from __future__ import print_function

from datetime import datetime
import time

from flask import current_app, url_for

from flask_user.tests.tst_utils import response_has_string



# **********************
# ** Global Variables **
# **********************
# Using global variable for speed
user1 = None
user2 = None
user3 = None
user4 = None

# *************
# ** Defines **
# *************
SHORT_USERNAME = 'Aa'
INVALID_EMAIL = 'user1.example.com'
invalid_usernames = (
    'with space',
    'with&symbol',
    "with'symbol",
    )
invalid_passwords = (
    'Abcd1',        # too short
    'ABCabc',       # no digits
    'ABC123',       # no lower case letters
    'abc123',       # no upper case letters
    )


# *********************
# ** Automated tests **
# *********************
# Function names must start with 'test'
# The 'client' parameter is set up in conftest.py

def test_init(db):
    """
    Set up two test users
    """
    global user1, user2, user3, user4

    # Enable all features
    um =  current_app.user_manager
    um.enable_register = True
    um.enable_change_username = True
    um.enable_change_password = True
    um.enable_confirm_email = True
    um.enable_reset_password = True
    um.enable_email = True
    um.enable_retype_password = True

    # Tests have not been written with auto_login in mind
    um.auto_login = False

    hashed_password = um.hash_password('Password1')
    User = um.db_adapter.UserClass

    # Create user1 with username and email
    user1 = User(username='user1', email='user1@example.com', password=hashed_password, active=True)
    assert user1
    db.session.add(user1)

    # Create user1 with email only
    user2 = User(email='user2@example.com', password=hashed_password, active=True)
    assert user2
    db.session.add(user2)

    # Create user3 with username and email
    user3 = User(username='user3', email='user3@example.com', password=hashed_password, active=True)
    assert user3
    db.session.add(user3)

    # Create user4 with email only
    user4 = User(email='user4@example.com', password=hashed_password, active=True)
    assert user4
    db.session.add(user4)

    db.session.commit()


def test_invalid_register_with_username_form(client):
    print("test_invalid_register_with_username_form")

    # Choose config
    um =  current_app.user_manager
    um.enable_username = True
    User = um.db_adapter.UserClass

    # Set default values
    url = url_for('user.register')
    username = 'user3'
    email = 'user3@example.com'
    password = 'Password1'

    # Test empty username
    client.post_invalid_form(url, 'Username is required',
            username='', email=email, password=password, retype_password=password)

    # Test short username
    client.post_invalid_form(url, 'Username must be at least 3 characters long',
            username=SHORT_USERNAME, email=email, password=password, retype_password=password)

    # Test invalid usernames
    for invalid_username in invalid_usernames:
        client.post_invalid_form(url, 'Username may only contain letters, numbers, ',
                username=invalid_username, email=email, password=password, retype_password=password)

    # Test existing username (case INsensitive!)
    client.post_invalid_form(url, 'This Username is already in use. Please try another one.',
            username='UsEr1', email=email, password=password, retype_password=password)

    # Test empty password
    client.post_invalid_form(url, 'Password is required',
            username=username, email=email, password='', retype_password='')

    # Test invalid passwords
    for invalid_password in invalid_passwords:
        client.post_invalid_form(url, 'Password must have at least 6 characters with one lowercase letter, one uppercase letter and one number',
                username=username, email=email, password=invalid_password, retype_password=invalid_password)

    # Test non-matching passwords
    client.post_invalid_form(url, 'Password and Retype Password did not match',
            username=username, email=email, password='Password1', retype_password='Password9')

def test_invalid_register_with_email_form(client):
    print("test_invalid_register_with_email_form")

    # Choose config
    um =  current_app.user_manager
    um.enable_username = False
    User = um.db_adapter.UserClass

    # Set default values
    url = url_for('user.register')
    email = 'user3@example.com'
    password = 'Password1'

    # Test empty email
    client.post_invalid_form(url, 'Email is required',
            email='', password=password, retype_password=password)

    # Test invalid email
    client.post_invalid_form(url, 'Invalid Email',
            email=INVALID_EMAIL, password=password, retype_password=password)

    # Test existing email (case INsensitive!)
    # TODO: Debug
    #client.post_invalid_form(url, 'This Email is already in use. Please try another one.',
    #        email='UsEr1@ExAmPlE.CoM', password=password, retype_password=password)

    # Test empty password
    client.post_invalid_form(url, 'Password is required',
            email=email, password='', retype_password='')

    # Test invalid passwords
    for invalid_password in invalid_passwords:
        client.post_invalid_form(url, 'Password must have at least 6 characters with one lowercase letter, one uppercase letter and one number',
                email=email, password=invalid_password, retype_password=invalid_password)

    # Test non-matching passwords
    client.post_invalid_form(url, 'Password and Retype Password did not match',
            email=email, password='Password1', retype_password='Password9')

def test_invalid_confirm_email_page(client):
    print("test_invalid_confirm_email_page")

    # Test Invalid token
    url = url_for('user.confirm_email', token='InvalidToken')
    client.get_invalid_page(url, 'Invalid confirmation token')

    # Generate valid token
    um = current_app.user_manager
    token = um.generate_token(user1.id)
    url = url_for('user.confirm_email', token=token)

    # Test Expired token
    um.confirm_email_expiration = 1   # set 1 second expiration
    time.sleep(2)                     # wait for 2 seconds
    client.get_invalid_page(url, 'Your confirmation token has expired')


def test_invalid_login_with_username_form(client):
    print("test_invalid_login_with_username_form")

    # Choose config
    um = current_app.user_manager
    um.enable_email = True
    um.enable_username = True

    # Set default values
    url = url_for('user.login')
    username = 'user1'
    password = 'Password1'

    # Test empty username
    client.post_invalid_form(url, 'Username is required',
            username='', password=password)

    # Test incorrect username
    um.show_username_email_does_not_exist = False
    client.post_invalid_form(url, 'Incorrect Username/Email and/or Password',
            username='Xuser1', password=password)
    um.show_username_email_does_not_exist = True
    client.post_invalid_form(url, 'Username/Email does not exist',
            username='Xuser1', password=password)
    um.show_username_email_does_not_exist = False

    # Test empty password
    client.post_invalid_form(url, 'Password is required',
            username=username, password='')

    # Test incorrect password
    um.show_username_email_does_not_exist = False
    client.post_invalid_form(url, 'Incorrect Username/Email and/or Password',
            username=username, password='XPassword1')
    um.show_username_email_does_not_exist = True
    client.post_invalid_form(url, 'Incorrect Password',
            username=username, password='XPassword1')
    um.show_username_email_does_not_exist = False

def test_invalid_login_with_email_form(client):
    print("test_invalid_login_with_email_form")

    # Choose config
    um = current_app.user_manager
    um.enable_email = True
    um.enable_username = False

    # Set default values
    url = url_for('user.login')
    email = 'user2@example.com'
    password = 'Password1'

    # Test empty email
    client.post_invalid_form(url, 'Email is required',
            email='', password=password)

    # Test incorrect email
    um.show_username_email_does_not_exist = False
    client.post_invalid_form(url, 'Incorrect Email and/or Password',
            email='Xuser2@example.com', password=password)
    um.show_username_email_does_not_exist = True
    client.post_invalid_form(url, 'Email does not exist',
            email='Xuser2@example.com', password=password)
    um.show_username_email_does_not_exist = False

    # Test empty password
    client.post_invalid_form(url, 'Password is required',
            email=email, password='')

    # Test incorrect password
    um.show_username_email_does_not_exist = False
    client.post_invalid_form(url, 'Incorrect Email and/or Password',
            email=email, password='XPassword1')
    um.show_username_email_does_not_exist = True
    client.post_invalid_form(url, 'Incorrect Password',
            email=email, password='XPassword1')
    um.show_username_email_does_not_exist = False

def test_invalid_change_username_form(client):
    print("test_invalid_change_username_form")

    # Set user manager config
    um =  current_app.user_manager
    um.enable_username = True
    um.enable_email = False

    # Set default values
    username = 'user1'
    password = 'Password1'
    new_username = 'user4'
    url = url_for('user.change_username')

    # Log in as 'user1'
    client.login(username=username, password=password)

    # Test empty username
    client.post_invalid_form(url, 'Username is required',
            new_username='', old_password=password)

    # Test short username
    client.post_invalid_form(url, 'Username must be at least 3 characters long',
            new_username=SHORT_USERNAME, old_password=password)

    # Test existing username
    client.post_invalid_form(url, 'This Username is already in use. Please try another one.',
            new_username='user3', old_password=password)

    # Test empty password
    client.post_invalid_form(url, 'Old Password is required',
            new_username=username, old_password='')

    # Test incorrect password
    client.post_invalid_form(url, 'Old Password is incorrect',
            new_username=username, old_password='XPassword1')

    client.logout()

def test_invalid_change_password_form(client):
    print("test_invalid_change_password_form")

    # Set user manager config
    um =  current_app.user_manager
    um.enable_username = False

    # Set default values
    email = 'user2@example.com'
    old_password = 'Password1'
    new_password = 'Password5'
    url = url_for('user.change_password')

    # Log in as 'user1'
    client.login(email=email, password=old_password)

    # Test empty old password
    client.post_invalid_form(url, 'Old Password is required',
            old_password='', new_password=new_password, retype_password=new_password)

    # Test incorrect old password
    client.post_invalid_form(url, 'Old Password is incorrect',
            old_password='XPassword1', new_password=new_password, retype_password=new_password)

    # Test empty password
    client.post_invalid_form(url, 'New Password is required',
            old_password=old_password, new_password='', retype_password=new_password)

    # Test invalid passwords
    for invalid_password in invalid_passwords:
        client.post_invalid_form(url, 'Password must have at least 6 characters with one lowercase letter, one uppercase letter and one number',
            old_password=old_password, new_password=invalid_password, retype_password=new_password)

    # Test non-matching passwords
    client.post_invalid_form(url, 'New Password and Retype Password did not match',
            old_password=old_password, new_password=new_password, retype_password='Xpassword5')

    client.logout()


def test_invalid_forgot_password_form(client):
    print("test_invalid_forgot_password_form")

    url = url_for('user.forgot_password')

    # Test invalid email
    client.post_invalid_form(url, 'Invalid Email',
            email=INVALID_EMAIL)


def test_invalid_reset_password(client):
    print("test_invalid_reset_password")

    # Set user manager config
    um =  current_app.user_manager

    # Set default values
    new_password = 'Password5'
    # Simulate a valid forgot password form
    token = um.generate_token(user1.id)

    # Test invalid token
    url = url_for('user.reset_password', token='InvalidToken')
    client.post_invalid_form(url, 'Your reset password token is invalid',
            new_password=new_password, retype_password=new_password)

    # Expired Token
    url = url_for('user.reset_password', token=token)
    um.reset_password_expiration = 1    # set 1 second expiration
    time.sleep(2)                       # wait for 2 seconds
    client.post_invalid_form(url, 'Your reset password token has expired',
            new_password=new_password, retype_password=new_password)
    um.reset_password_expiration = 2*24*3600  # 2 days

    # Invalid retype password
    client.post_invalid_form(url, 'New Password and Retype Password did not match',
            new_password = new_password, retype_password='XPassword5')

def test_valid_roles(client):
    um =  current_app.user_manager
    User = um.db_adapter.UserClass

    # Perform only for roles_required_app
    user007 = User.query.filter(User.username=='user007').first()
    if not user007: return

    print("test_valid_roles")
    um.enable_username = True

    client.login(username='user007', password='Password1')
    url = url_for('special_page')
    response = client.get_valid_page(url)
    assert not response_has_string(response, 'You must be signed in to access')
    client.logout()

def test_invalid_roles(client):
    um =  current_app.user_manager
    User = um.db_adapter.UserClass

    # Perform only for roles_required_app
    user007 = User.query.filter(User.username=='user007').first()
    if not user007: return

    print("test_invalid_roles")
    um.enable_username = True

    client.login(username='user1', password='Password1')
    url = url_for('special_page')
    response = client.get_invalid_page(url, 'You do not have permission to access')
    client.logout()

def test_login_without_confirm_email(client):
    print("test_login_without_confirm_email")

    um = current_app.user_manager
    um.enable_username = False
    um.enable_email = True
    um.enable_confirm_email = True
    um.enable_retype_password = False

    email = 'notconfirmed@example.com'
    password = 'Password1'

    # register user
    client.post_valid_form(url_for('user.register'),
            email=email,
            password=password)

    # Try logging in without confirming email
    client.post_invalid_form(url_for('user.login'),
            'Your email address has not yet been confirmed',
            email=email,
            password=password)

    # Confirm email manually, but disable account
    User = um.db_adapter.UserClass
    user = User.query.filter(User.email==email).first()
    assert(user)
    user.active = False
    user.confirmed_at = datetime.utcnow()

    # Try logging in into  disabled account
    client.post_invalid_form(url_for('user.login'),
            'Your account has not been enabled',
            email=email,
            password=password)

def test_cleanup(db):
    """
    Delete user1 and user2
    """
    global user1, user2, user3, user4
    db.session.delete(user1)
    db.session.delete(user2)
    db.session.delete(user3)
    db.session.delete(user4)
    db.session.commit()
    user1 = None
    user2 = None
    user3 = None
    user4 = None


# TODO:
# Register without confirming email and try to log in
# 'Your email address has not yet been confirmed. Check your email Inbox and Spam folders for the confirmation email and follow the instructions to activate your account.'
#
# Disable account and try to login
# 'Your account has been disabled.'
#
# Logout with user_manager.logout_next set
#
# Reset password with custom user_manager.password_validator
#
# Change password with custom user_manager.password_validator:
#
# Custom db_adapter.EmailClass


"""Database models and connectivity."""
from . import __VERSION__
from flask_sqlalchemy import SQLAlchemy
from .flask_user import UserMixin
from pkg_resources import packaging
from sqlalchemy.sql import func
import logging
from sqlalchemy.ext.orderinglist import ordering_list

db = SQLAlchemy()

class Annotation(db.Model):
    __tablename__ = 'annotations'
    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(db.Integer, db.ForeignKey('users.id', ondelete='RESTRICT'), nullable=False)
    project_id = db.Column(db.Integer, db.ForeignKey('projects.id', ondelete='RESTRICT'), nullable=False)
    file_id = db.Column(db.Integer, db.ForeignKey('files.id', ondelete='RESTRICT'), nullable=False)
    pattern_id = db.Column(db.Integer, db.ForeignKey('patterns.id', ondelete='RESTRICT'), nullable=True)
    pattern_set_id = db.Column(db.Integer, db.ForeignKey('pattern_sets.id', ondelete='RESTRICT'), nullable=True)
    series = db.Column(db.String(255), nullable=False)
    left = db.Column(db.Float, nullable=True)
    right = db.Column(db.Float, nullable=True)
    top = db.Column(db.Float, nullable=True)
    bottom = db.Column(db.Float, nullable=True)
    label = db.Column(db.Text, nullable=False)
    created_at = db.Column(db.DateTime, nullable=False, server_default=func.now())

    file = db.relationship('File', lazy='joined', backref=db.backref('annotations', lazy=True))
    user = db.relationship('User', lazy=True)

    def __repr__(self):
        return f"ID: {self.id}, UID: {self.user_id}, PID: {self.project_id}, FID: {self.file_id}, PID: {self.pattern_id}, Series: {self.series}, Boundaries: {self.left} {self.right} {self.top} {self.bottom}, Label: {self.label}"

class AnnotationTemplate(db.Model):
    __tablename__ = 'annotation_templates'
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(255), nullable=False)
    form = db.Column(db.Text, nullable=False)
    created_at = db.Column(db.DateTime, nullable=False, server_default=func.now())

class Pattern(db.Model):
    __tablename__ = 'patterns'
    id = db.Column(db.Integer, primary_key=True)
    pattern_set_id = db.Column(db.Integer, db.ForeignKey('pattern_sets.id', ondelete='CASCADE'), nullable=False)
    project_id = db.Column(db.Integer, db.ForeignKey('projects.id', ondelete='RESTRICT'), nullable=False)
    file_id = db.Column(db.Integer, db.ForeignKey('files.id', ondelete='CASCADE'), nullable=False)
    series = db.Column(db.String(255), nullable=False)
    left = db.Column(db.Float, nullable=True)
    right = db.Column(db.Float, nullable=True)
    top = db.Column(db.Float, nullable=True)
    bottom = db.Column(db.Float, nullable=True)
    label = db.Column(db.Text, nullable=True)
    created_at = db.Column(db.DateTime, nullable=False, server_default=func.now())

    annotations = db.relationship('Annotation', lazy=True, backref=db.backref('pattern', lazy=True))
    file = db.relationship('File', lazy='joined', backref=db.backref('patterns', lazy=True))

    def __repr__(self):
        return f"ID: {self.id}, PSID: {self.pattern_set_id}, FID: {self.file_id}, Series: {self.series}, Boundaries: {self.left} {self.right} {self.top} {self.bottom}, Label: {self.label}" + (f"\n    Annotations: {self.annotations}" if len(self.annotations)>0 else "")


# Many-to-many table joining pattern sets to users vis-a-vis assignments
patternSetAssignments = db.Table('pattern_set_assignments',
    db.Column('pattern_set_id', db.Integer, db.ForeignKey('pattern_sets.id'), primary_key=True),
    db.Column('user_id', db.Integer, db.ForeignKey('users.id'), primary_key=True)
)


class PatternSet(db.Model):
    __tablename__ = 'pattern_sets'
    id = db.Column(db.Integer, primary_key=True)
    project_id = db.Column(db.Integer, db.ForeignKey('projects.id', ondelete='CASCADE'), nullable=False)
    #annotation_template_id = db.Column(db.Integer, db.ForeignKey('annotation_templates.id', ondelete='RESTRICT'), nullable=True)
    name = db.Column(db.String(255), nullable=False)
    description = db.Column(db.String(255), nullable=True)
    show_by_default = db.Column(db.Boolean, default=False, nullable=False)
    created_at = db.Column(db.DateTime, nullable=False, server_default=func.now())

    annotations = db.relationship('Annotation', lazy=True)
    patterns = db.relationship('Pattern', lazy=True)
    project = db.relationship('Project', lazy='joined', backref=db.backref('pattern_sets', lazy=True))
    users = db.relationship('User', secondary=patternSetAssignments, lazy=True, backref=db.backref('pattern_sets', lazy=True))


class File(db.Model):
    __tablename__ = 'files'
    id = db.Column(db.Integer, primary_key=True)
    project_id = db.Column(db.Integer, db.ForeignKey('projects.id', ondelete='CASCADE'), nullable=False)
    #project = db.relationship('Project', backref=db.backref('files', lazy=True))
    path = db.Column(db.String(255), nullable=False)


class Project(db.Model):
    __tablename__ = 'projects'
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(255), nullable=False)
    path = db.Column(db.String(255), nullable=False)


class Role(db.Model):
    __tablename__ = 'roles'
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(50), nullable=False, server_default=u'', unique=True)  # for @roles_accepted()


class User(db.Model, UserMixin):
    __tablename__ = 'users'

    id = db.Column(db.Integer, primary_key=True)

    # User authentication information. The collation='NOCASE' is required
    # to search case insensitively when USER_IFIND_MODE is 'nocase_collation'.
    email = db.Column(db.String(255, collation='NOCASE'), nullable=False, unique=True)
    confirmed_at = db.Column(db.DateTime(), nullable=True)
    password = db.Column(db.String(255), nullable=False, server_default='')

    # User information
    active = db.Column('is_active', db.Boolean(), nullable=False, server_default='0')
    first_name = db.Column(db.Unicode(50), nullable=False, server_default=u'')
    last_name = db.Column(db.Unicode(50), nullable=False, server_default=u'')

    # Relationships
    roles = db.relationship('Role', secondary='users_roles', backref=db.backref('users', lazy='dynamic'))

    # Other
    assignments_remaining = db.Column(db.Integer, nullable=False, default=0)


class UserInvitation(db.Model):
    __tablename__ = 'user_invite'
    id = db.Column(db.Integer, primary_key=True)
    email = db.Column(db.String(255), nullable=False)
    # save the user of the invitee
    invited_by_user_id = db.Column(db.Integer, db.ForeignKey('users.id'))
    # token used for registration page to identify user registering
    token = db.Column(db.String(100), nullable=False, server_default='')


# The version table holds a single row with a single column and stores
# the auviewer version which the database matches.
class Version(db.Model):
    __tablename__ = 'version'
    id = db.Column(db.Integer, primary_key=True)
    version = db.Column(db.String(255), nullable=False)


# Define the UserRoles association model
class UsersRoles(db.Model):
    __tablename__ = 'users_roles'
    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(db.Integer, db.ForeignKey('users.id', ondelete='CASCADE'))
    role_id = db.Column(db.Integer, db.ForeignKey('roles.id', ondelete='CASCADE'))


def init_flask_app(app):

    db.init_app(app)

    # Create all database tables
    # TODO(gus): Once we remove flask-user and detach this from flask, move this.
    with app.app_context():

        # Create database & tables, if needed
        db.create_all()

        # Perform db upgrades, if needed
        upgrade_db()

    # app.app_context().push()
    # db.create_all()

# Perform any db version upgrades necessary
def upgrade_db():

    # Get auviewer package version
    app_version = packaging.version.parse(__VERSION__)

    # Check application version, and upgrade iteratively as needed
    while True:

        # Get db version
        dvm = Version.query.first()

        # If there is no version, assume 0.1.1rc4 (after this, db
        # versioning began) and write this to the database.
        if dvm is None:
            db.session.add(Version(id=0, version='0.1.1rc4'))
            db.session.commit()
            continue

        db_version = packaging.version.parse(dvm.version)

        # ADD DB UPGRADE STEPS LIKE THIS:
        # if db_version < packaging.version.parse('x.x.x'):
        #     do_some_db_upgrade()
        #     Version.query.first().version = 'x.x.x'
        #     db.session.commit()
        #     continue

        if db_version != app_version:
            logging.info(f"Database version was {db_version}. Upgrading to {app_version}. __VERSION__ is {__VERSION__}.")
            Version.query.first().version = __VERSION__
            db.session.commit()

        # We've completed all db upgrades, so break the loop
        break


from .model import SimpleFeaturizer, FeaturizerParameter


class MeanFeaturizer(SimpleFeaturizer):

    id = 'mean'
    name = 'Mean'
    parameters = [
        FeaturizerParameter(id='skipna', name="Skip NaN", description="Skip NaN values", data_type='boolean'),
    ]

    def featurize(self, data, params={}):
        skipna = params['skipna']
        return data.mean(skipna=skipna)


""" This file creates event notification signals for Flask-User.
    Signals are based on Flask.signals which are based on the blinker signals.

    :copyright: (c) 2013 by Ling Thio
    :author: Ling Thio (ling.thio@gmail.com)
    :license: Simplified BSD License, see LICENSE.txt for more details."""

from flask.signals import Namespace

_signals = Namespace()                              # Place Flask-User signals in our own namespace

# *******************
# ** Flask Signals **
# *******************
# Flask signals are based on blinker. Neither Flask nor Flask-User installs blinker
# If you plan to use signals, please install blinker with 'pip install blinker'
# See http://flask.pocoo.org/docs/signals/

# Sent when a user changed their password
user_changed_password = _signals.signal('user.user_changed_password')

# Sent when a user changed their username
user_changed_username = _signals.signal('user.user_changed_username')

# Sent when a user confirmed their email
user_confirmed_email = _signals.signal('user.user_confirmed_email')

# Sent when a user submitted a password reset request
user_forgot_password = _signals.signal('user.forgot_password')

# Sent when a user logged in
user_logged_in = _signals.signal('user.user_logged_in')

# Sent when a user logged out
user_logged_out = _signals.signal('user.user_logged_out')

# Sent when a user registered a new account
user_registered = _signals.signal('user.user_registered')

# Signal sent just after a password was reset
user_reset_password = _signals.signal('user.user_reset_password')

# Signal sent just after a user sent an invitation  # TODO: Not yet implemented
user_sent_invitation = _signals.signal('user.user_sent_invitation')




"""AUViewer package."""

__VERSION_MAJOR__ = '1'
__VERSION_MINOR__ = '0'
__VERSION_BUILD__ = '1'

__DEBUG = False
if __DEBUG:
    import datetime as dt
    __VERSION_TAGS__ = ''#f"d{dt.datetime.now().strftime('%Y%m%d_%H%M')}"
else:
    __VERSION_TAGS__ = ''

__VERSION__ = f'{__VERSION_MAJOR__}.{__VERSION_MINOR__}.{__VERSION_BUILD__}{__VERSION_TAGS__}'

# Set logging level to info
import logging
if __DEBUG:
    logging.basicConfig(level=logging.INFO)
    logging.getLogger('sqlalchemy.engine').setLevel(logging.INFO)
else:
    logging.basicConfig(level=logging.WARNING)
    logging.getLogger('sqlalchemy.engine').setLevel(logging.WARNING)



"""
    tests.test_all_forms.py
    ---------------------------
    Flask-User automated tests:
    Tests posting a sequence of valid forms for all possible config combinations

    :copyright: (c) 2013 by Ling Thio
    :author: Ling Thio (ling.thio@gmail.com)
    :license: Simplified BSD License, see LICENSE.txt for more details.
"""

from __future__ import print_function
import datetime

from flask import current_app, url_for

# **********************
# ** Global Variables **
# **********************
# Using global variable for speed
user1 = None

# ********************************
# ** Automatically called Tests **
# ********************************
# The 'client' and 'app' parameters are set up in conftest.py
# Functions that start with 'test' will be run automatically by the test suite runner (py.test)

def test_multiple_emails(app, db, client):
    """
    Test 'multiple emails per user' feature
    """

    # Set Flask-User settings
    um = current_app.user_manager
    um.enable_register = True
    um.enable_username = False
    um.enable_email = True
    um.enable_confirm_email = True
    um.enable_change_username = False
    um.enable_change_password = False
    um.enable_forgot_password = False
    um.enable_multiple_emails = True
    um.enable_retype_password = False

    # Adjust DbAdapter settings
    um.db_adapter.UserEmailClass = app.UserEmailClass

    # Adjust URL routes
    app.add_url_rule(um.email_action_url,  'user.email_action',  um.email_action_view_function)
    app.add_url_rule(um.manage_emails_url, 'user.manage_emails', um.manage_emails_view_function, methods=['GET', 'POST'])

    # constants
    EMAIL1 = 'email1@multi-email.com'
    EMAIL2 = 'email2@multi-email.com'
    PASSWORD = 'Password1'

    # Register user
    response = client.post_valid_form(url_for('user.register'), email=EMAIL1, password=PASSWORD)
    user_email1 = um.db_adapter.UserEmailClass.query.filter(um.db_adapter.UserEmailClass.email==EMAIL1).first()
    assert user_email1 != None

    # Confirm email
    confirmation_token = um.generate_token(user_email1.id)
    client.get_valid_page(url_for('user.confirm_email', token=confirmation_token))

    # Log in using email1
    client.login(email=EMAIL1, password=PASSWORD)

    # Visit manage emails page
    response = client.get_valid_page(url_for('user.manage_emails'))
    assert response.data.find(str.encode(EMAIL1)) >= 0

    # Add an email
    response = client.post_valid_form(url_for('user.manage_emails'), email=EMAIL2)
    assert response.data.find(str.encode(EMAIL1)) >= 0
    assert response.data.find(str.encode(EMAIL2)) >= 0
    user_email2 = um.db_adapter.UserEmailClass.query.filter(um.db_adapter.UserEmailClass.email==EMAIL2).first()
    assert user_email2 != None

    # Confirm email
    confirmation_token = um.generate_token(user_email2.id)
    client.get_valid_page(url_for('user.confirm_email', token=confirmation_token))

    # Logout
    client.logout()

    # Log in using email1
    client.login(email=EMAIL1, password=PASSWORD)

    # Logout
    client.logout()

    # Log in using email2
    client.login(email=EMAIL2, password=PASSWORD)

    # Confirm
    response = client.get_valid_page(url_for('user.email_action', id=user_email2.id, action='confirm'))

    # Make primary
    response = client.get_valid_page(url_for('user.email_action', id=user_email2.id, action='make-primary'))

    # Delete
    response = client.get_valid_page(url_for('user.email_action', id=user_email1.id, action='delete'))

    # Logout
    client.logout()

    # Restore settings
    um.enable_multiple_emails = False
    um.enable_confirm_email = True
    um.enable_retype_password = True
    um.db_adapter.UserEmailClass = None




"""Shared methods for the package."""

import pandas as pd
from pathlib import Path

# Creates an empty json file at the provided path (pathlib.Path object) if the
# file does not already exist.
def createEmptyJSONFile(path_obj):
    if not path_obj.exists():
        with path_obj.open(mode='x') as f:
            f.write("{\n\n}")

def annotationDataFrame(annotationModels):
    """Given a list of annotation models, returns a DataFrame in our standard format."""
    return pd.DataFrame(
        [[
            a.file.id,
            Path(a.file.path).name,
            a.user.id, a.user.email,
            a.user.first_name,
            a.user.last_name,
            a.pattern_set_id,
            a.pattern_id,
            a.series,
            a.left,
            a.right,
            a.top,
            a.bottom,
            a.label,
            a.created_at,
            f"{a.project_id}_{a.file.id}_{a.series}_{a.left}_{a.right}_{a.top}_{a.bottom}",
        ] for a in annotationModels],
        columns=['file_id', 'filename', 'user_id', 'user_email', 'user_firstname', 'user_lastname', 'pattern_set_id',
                 'pattern_id', 'series', 'left', 'right', 'top', 'bottom', 'label', 'created', 'pattern_identifier']
    )

# Given an Annotation or Pattern model, returns a list formatted for API output.
# A second parameter, related, may be provided in which case the function will
# output a representation for both items in a single list. For example, you may
# provide an annotation and the pattern it annotates, or a pattern and the
# annotation that annotates it.
# TODO: Review whether both cases will be used.
def annotationOrPatternOutput(primary, related=None):
    output = []
    for a in [primary, related]:
        if a is not None:
            output = output + [
                a.id,
                a.file_id,
                Path(a.file.path).name,
                a.series,
                a.left,
                a.right,
                a.top,
                a.bottom,
                a.label,
                a.pattern_id if hasattr(a, 'pattern_id') else None
            ]
    return output

def getProcFNFromOrigFN(fp):
    """Returns the processed filename (as Path object, which can be treated as string) from origina filename (as string or Path object)"""
    return Path(Path(fp).stem + '_processed.h5')

def patternDataFrame(patternModels):
    """Given a list of pattern models, returns a DataFrame in our standard format."""
    return pd.DataFrame(
            [[
                pattern.file.id,
                Path(pattern.file.path).name,
                pattern.series,
                pattern.left,
                pattern.right,
                pattern.top,
                pattern.bottom,
                pattern.label,
                f"{pattern.project_id}_{pattern.file.id}_{pattern.series}_{pattern.left}_{pattern.right}_{pattern.top}_{pattern.bottom}",
            ] for pattern in patternModels],
            columns=['file_id', 'filename', 'series', 'left', 'right', 'top', 'bottom', 'label', 'pattern_identifier']
        )


from .model import SimpleFeaturizer, FeaturizerParameter


class ExampleFeaturizer(SimpleFeaturizer):

    id = 'example'
    name = 'Example'
    parameters = [
        FeaturizerParameter(id='booleanr', name="Boolean Req'd", description="Boolean description", data_type='boolean', required=True),
        FeaturizerParameter(id='booleannr', name="Boolean Not Req'd", description="Boolean description", data_type='boolean', required=False),
        FeaturizerParameter(id='input', name="Input"),
        FeaturizerParameter(id='dropdown', name="Dropdown", description="Dropdown description", form_field_type='dropdown', options={'val1': 'Value 1', 'val2': 'Value 2'}),
        FeaturizerParameter(id='textarea', name='Textarea', description="Textarea description", form_field_type='textarea'),
        FeaturizerParameter(id='checkbox', name='Checkbox', description="Checkbox description", form_field_type='checkbox'),
        FeaturizerParameter(id='radio', name='Radio', description="Radio description", form_field_type='radio', options={'val1': 'Value 1', 'val2': 'Value 2', 'val3': 'Value 3'}),
        # FeaturizerParameter(id='', description="", data_type='slider'),
    ]
    neededSeries = 2

    def featurize(self, data, params):
        print(f"Parameters received:\n{params}")
        return data.mean()